package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.io.*;
import java.lang.*;
import java.text.*;
import java.util.Vector;

/**
 * A BoundedFeatureDescriptor contains bounds on the values which it
 * might contain.  These bounds are represented as Number object, and
 * are compared as Doubles.
 * <P>
 * $Id: BoundedFeatureDescriptor.java,v 1.7 1998/10/16 01:56:59 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       IntegralFeatureDescriptor
 * @see       ContinuousFeatureDescriptor
 * @version   $Revision: 1.7 $
 **/
public abstract class BoundedFeatureDescriptor
extends FeatureDescriptor
{
  /**
   * serialization Stream Unique Identifier.  Recompute [with
   * serialver] when this class changes.
   **/
  static final long serialVersionUID = -5436088333921205272L;


  /* ATTRIBUTES */
  
  /** The lower bound of value range. */
  protected Number lowerBound;

  /** The upper bound of value range. */
  protected Number upperBound;

  /* CONSTRUCTORS */

  public BoundedFeatureDescriptor(String newName,
				  Number lowerBound,
				  Number upperBound)
  {
    super(newName);
    setBounds(lowerBound, upperBound);
  }

  public BoundedFeatureDescriptor(String newName,
				  Number lowerBound,
				  Number upperBound,
				  ProbDist newDefaultDist)
  {
    super(newName, newDefaultDist);
    setBounds(lowerBound, upperBound);
  }

  /* METHODS */

  /** Set a new upper bound on the feature value range.
    Makes sure that lower bound is not higher than upper bound.
    @param newUpperBound int value of new upper bound for 
    feature value range
    **/
   protected void setUpperBound(Number newUpperBound)
   {
      upperBound = newUpperBound;
      if ( (lowerBound != null) &&
	   (lowerBound.doubleValue() > upperBound.doubleValue()) ) {
	lowerBound = upperBound;
      }
   }

   /** Set a new lower bound on the feature value range.
       Makes sure that upper bound is not lower than lower bound.
       @param newLowerBound value of new lower bound for 
                            feature value range */
   protected void setLowerBound(Number newLowerBound)
   {
      lowerBound = newLowerBound;
      if ( (upperBound != null) &&
	   (upperBound.doubleValue() < lowerBound.doubleValue()) ) {
         upperBound = lowerBound; 
      }
   }

   /** Set new bounds on the feature value range.
       Makes sure that lower bound is not higher than upper bound.
       @param newLowerBound value of new lower bound for 
                            feature value range
       @param newUpperBound value of new upper bound for 
                            feature value range */
   protected void setBounds(Number newLowerBound, Number newUpperBound)
   {
      setLowerBound(newLowerBound);
      setUpperBound(newUpperBound);
   }

   /** Get the lower bound of the feature value range.
       @return returns the lower bound of the feature value range */
   public Number getLowerBound() 
   {
      return lowerBound;
   }

   /** Get the upper bound of the feature value range.
       @return returns the upper bound of the feature value range */
   public Number getUpperBound()
   {
      return upperBound;
   } 

  /**
   * Redefinition of Object.clone()
   **/
  public Object clone()
  {
    BoundedFeatureDescriptor toRet = (BoundedFeatureDescriptor)super.clone();
    if (upperBound instanceof Double) {
      toRet.upperBound = new Double( upperBound.doubleValue() );
      toRet.lowerBound = new Double( lowerBound.doubleValue() );
    } else {
      toRet.upperBound = new Integer( upperBound.intValue() );
      toRet.lowerBound = new Integer( lowerBound.intValue() );
    }
    return toRet;
  }

  /**
   * Redefinition of Object.equals(Object)
   **/
  public boolean equals(Object o)
  {
    if (!super.equals(o)) { return false; }
    if (! getClass().isInstance(o)) { return false; }
    BoundedFeatureDescriptor bfd = (BoundedFeatureDescriptor)o;
    if (! (bfd.upperBound.equals(upperBound) &&
	   bfd.lowerBound.equals(lowerBound)) ) {
      return false;
    }
    return true;
  }

}

/* INNER CLASSES for EDITING */
abstract class BFDViewer
    extends FDViewer
{
    /**
     * The JTextFields used for the upper/lower bound editing
     **/
    JTextField lowerBoundTF;
    JTextField upperBoundTF;
    /**
     * The NumberFormat object to use for number formatting/parsing.
     **/
    NumberFormat nf;
    
    public BFDViewer(FeatureDescriptor this_fd, boolean edit)
    {
	super(this_fd, edit);
    }

    public void createOptionalPanel()
    {
	if (nf == null) { nf = NumberFormat.getNumberInstance(); }

	JLabel lbl, ubl;
	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();

	optPanel = new JPanel(gbl);

	BoundedFeatureDescriptor bfd = (BoundedFeatureDescriptor)fd;

	lowerBoundTF = new JTextField( nf.format(bfd.getLowerBound()), 10 );
	upperBoundTF = new JTextField( nf.format(bfd.getUpperBound()), 10 );

	if (viewing) {
	    lowerBoundTF.setEditable(false);
	    upperBoundTF.setEditable(false);
	}

	gbc.fill = GridBagConstraints.NONE;
	gbc.gridwidth = GridBagConstraints.RELATIVE;
	lbl = new JLabel("Lower Bound:");
	ubl = new JLabel("Upper Bound:");
	gbl.setConstraints(lbl, gbc);
	gbl.setConstraints(ubl, gbc);

	gbc.fill = GridBagConstraints.HORIZONTAL;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(lowerBoundTF, gbc);
	gbl.setConstraints(upperBoundTF, gbc);

	optPanel.add(lbl);
	optPanel.add(lowerBoundTF);
	optPanel.add(ubl);
	optPanel.add(upperBoundTF);

    }

    public abstract void createProbPanel();

    public abstract void createDetPanel();

    public void canSaveValues(Vector reasons)
    {
	super.canSaveValues(reasons);
      
	try {
	    nf.parse(lowerBoundTF.getText());
	} catch (ParseException pe) {
	    reasons.addElement("Cannot parse lower bound " +
			       lowerBoundTF.getText() +
			       " [" +
			       pe.getMessage() +
			       "].");
	} // end try-catch for lower-bound text field.

	try {
	    nf.parse(upperBoundTF.getText());
	} catch (ParseException pe) {
	    reasons.addElement("Cannot parse upper bound " +
			       upperBoundTF.getText() +
			       " [" +
			       pe.getMessage() +
			       "].");
	} // end try-catch for upper-bound text field.
    }

    public void saveValues()
	throws CannotSaveValuesException
    {
	if (viewing) {
	    return;
	}
	super.saveValues();
	try {
	    BoundedFeatureDescriptor bfd = (BoundedFeatureDescriptor)fd;
	    // Save the lower bound
	    bfd.setLowerBound( nf.parse(lowerBoundTF.getText()) );
	    // Save the upper bound
	    bfd.setUpperBound( nf.parse(upperBoundTF.getText()) );
	} catch (ParseException pe) {
	    throw new CannotSaveValuesException( pe.getMessage() );
	} // end try-catch
    }

}
