package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.*;
import java.text.*;
import java.util.Vector;
import NTLSystem.NTLException;

/**
   A continuous feature-descriptor contains meta information that pertains
   to all instantiations of a particular continuous feature.
   <P>
   $Id: ContinuousFeatureDescriptor.java,v 1.20 1998/10/16 01:57:00 jsled Exp $
   <BR>
   Last editor: $Author: jsled $

   @author     Tajh Taylor [tajh@icsi.berkeley.edu]
   @version    $Revision: 1.20 $
 **/

public class ContinuousFeatureDescriptor
extends BoundedFeatureDescriptor
{
  /**
   * serialization Stream Unique Identifier.  Recompute [with
   * serialver] when this class changes.
   **/
  static final long serialVersionUID = 6096012845341696798L;

  /* ATTRIBUTES */

  // public instance methods

   // constructors

   /** A no-argument constructor for serialization. */
   public ContinuousFeatureDescriptor()
   {
     super("",
	   new Double(Double.NEGATIVE_INFINITY),
	   new Double(Double.POSITIVE_INFINITY));
     setDefaultValue(new Double(0.0));
   }

   /** A basic constructor for determinate continuous features.
       Assumes bounds are infinite and default value is 0.0.
       @param newName the new name for the feature */
   public ContinuousFeatureDescriptor(String newName)
   {
      super(newName,
	    new Double(Double.NEGATIVE_INFINITY),
	    new Double(Double.POSITIVE_INFINITY));
      setDefaultValue(new Double(0.0));
   }

   /** A basic constructor for probabilistic continuous features.
       Assumes bounds are infinite and the default value is 0.0.
       @param newName the new name for the feature
       @param newDefaultDist the new default distribution */
   public ContinuousFeatureDescriptor(String newName, 
				      ProbDist newDefaultDist)
   {
      super(newName, 
	    new Double(Double.NEGATIVE_INFINITY),
	    new Double(Double.POSITIVE_INFINITY),
	    newDefaultDist);
   }

   /** A detailed constructor for determinate continuous features.
       @param     newName the new name for the feature
       @param     newLowerBound the new lower bound on feature values
       @param     newUpperBound the new upper bound on feature values
       @param     newDefault the new default value
       @exception NTLException If setDefaultValue throws it.
       */
   public ContinuousFeatureDescriptor(String newName,
				      double newLowerBound,
				      double newUpperBound,
				      double newDefault)
      throws NTLException
   {
      super(newName,
	    new Double(newLowerBound),
	    new Double(newUpperBound));
      setDefaultValue(new Double(newDefault));
   } 

   /** A detailed constructor for probabilistic continuous features.
       @param newName the new name for the feature
       @param newLowerBound  the new lower bound on feature values
       @param newUpperBound  the new upper bound on feature values
       @param newDefaultDist the new default distribution */
   public ContinuousFeatureDescriptor(String newName,
				      double newLowerBound, 
				      double newUpperBound,
				      ProbDist newDefaultDist) 
  {
      super(newName,
	    new Double(newLowerBound), new Double(newUpperBound),
	    newDefaultDist);
   } 

   // implementation of abstract FeatureDescriptor methods

   /** tell whether or not we are ordered; always true.
       @return always returns true */
   public boolean isOrdered()
   {
      return true;
   }

  /** Test whether the argument object contains a valid value for
       this particular feature.  Note that implementations should 
       report false for ProbDist object if the feature is determinate,
       and for definite value object if the feature is probabilistic.
       @param o any object to be tested.
       @return boolean flag indicating validity. */
   public boolean isValidValue(Object o)
   {
      Double d;
      ProbDist pd;
      
      // check and see if the feature is probabilistic first
      if (isProbabilistic())
      {
	 // make sure the distribution is probabilistic
	 if (o instanceof ProbDist)
	 {
	    pd = (ProbDist)o;
	    if (pd.isContinuous())
	       return true;
	    else
	       return false;
	 }
	 else
	    return false;
      }
      else // it's a determinate feature
      {
	 // make sure it's not a ProbDist
	 if (o instanceof ProbDist)
	    return false;
	 
	 // make sure it IS a double wrapped in a Double
	 if (o instanceof Double)
	 {
	    d = (Double)o;

	    // check the bounds
	    if ((d.doubleValue() >= getLowerBound().doubleValue()) &&
		(d.doubleValue() <= getUpperBound().doubleValue()))
	       return true;
	    else
	       return true;
	 }
	 else
	    return false;
      }
   }

  /** Return the value with the maximum likelihood,
      given a param.
      @param     the distribution to draw from
      @return    the ML value as a Double
      @exception NTLException If the given distribution is not
                 Continuous.
      */
   public Object getMLValue(ProbDist distribution)
      throws NTLException
   {
      Double d;
      
      if (!(distribution.isContinuous())) {
	  throw new NTLException("Must use continuous distribution " +
				 "to generate ML value for continuous " +
				 "feature.");
      }

      // FIXME: Add getMLValue to ProbDist interface.
      // d = (Double)distribution.getMLValue();
      // return d;
      return new Double(0.0);
   }

  /** Return a value at random according to the distribution.
      @param     distribution to draw from
      @return    the chosen value as a Double
      @exception NTLException If the distribution is not continuous.
      */
   public Object chooseValue(ProbDist distribution)
       throws NTLException
   {
       Double d;
      
       if (!(distribution.isContinuous())) {
	  throw new NTLException("Must use continuous distribution " +
				 "to generate random value for continuous " +
				 "feature.");
       }

       // FIXME: Add getMLValue to ProbDist interface.
       //d = (Double)distribution.getMLValue();
       //return d;
       return new Double(0.0);
   }
  
   // class-specific methods

   /** Set the default value for this feature.
       Makes the feature determinate.
       @param newDefault the new default value for this feature */
    protected void setDefaultValue(Double newDefault)
    {
	defaultValue = newDefault;
	setProbabilistic(false);
    }
   

  /**
   * @see FeatureDescriptor
   **/
  public Object valueFromText(String textVal)
       throws IllegalArgumentException
  {
    Double toRet = null;
    NumberFormat nf = NumberFormat.getNumberInstance();
    try {
      toRet = new Double((nf.parse(textVal)).doubleValue());
    } catch (ParseException pe) {
      throw new IllegalArgumentException(pe.getMessage());
    }
    if (toRet == null) {
      throw new IllegalArgumentException("No number parsed from" + textVal);
    }
    return toRet;
  }

  /**
   * @see FeatureDescriptor
   **/
  public String valueAsText(Object val)
  {
    NumberFormat nf = NumberFormat.getNumberInstance();
    return nf.format( (Double)val );
  }

   /** Test for equality with another object.
       @param o an object to be tested
       @return boolean flag indicating equality */
   public boolean equals(Object o)
   {
     if (!super.equals(o)) { return false; }
     ContinuousFeatureDescriptor cfd = (ContinuousFeatureDescriptor)o;
     if (! getDefaultValue().equals(cfd.getDefaultValue())) {
       return false;
     }
     return true;
   }

  /* implementation of Externalizable methods */

   /** Read in the descriptor from human-readable text.
       @param input input source */
   public void readExternal(ObjectInput input)
      throws IOException
   {
      //TODO requires a generic word parser class
   }

   /** Write the descriptor out in human-readable text form.
       Completely overrides the parent method.
       @param output output sink */
   public void writeExternal(ObjectOutput output)
       throws IOException
   {
      // write a header
      output.writeChars("ContinuousFeatureDescriptor\n");
   }

  public String getDisplayableName()
  {
    return new String("Continuous Feature Descriptor");
  }

  /**
   * Since editing and viewing panels are so similar, they can be
   * generated w/in the same method.
   *
   * @param     edit True if editing ability is requested, false if
   *            viewing only. 
   **/
  protected JPanel getEditViewPanel(boolean edit)
  {
    viewer = new CFDViewer(this, edit);
    return viewer;
  }

}

/* INNER CLASS for EDITING */
class CFDViewer
    extends BFDViewer
{
    /**
     * The JTextField used for default value editing.
     **/
    JTextField defValTF;

    /**
     * The new ProbDist object [when editing].
     **/
    ProbDist defPD;
    /**
     * The ComboBox to use for selecting which type of Distribution to
     * create.
     **/
    JComboBox distCombo;
    /**
     * The sub-panel actually containing the ProbabilityDistribution editor.
     **/
    JPanel subProbPanel;

    public CFDViewer(FeatureDescriptor this_fd, boolean edit)
    {
	super(this_fd, edit);
    }

    public void createProbPanel()
    {
	// Create the ProbDist-creation panel.
	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	probPanel = new JPanel(gbl);
	subProbPanel = new JPanel();

	gbc.fill = GridBagConstraints.NONE;
	gbc.gridwidth = GridBagConstraints.RELATIVE;

	// Present the option to create a distribution.
	distCombo = new JComboBox();
	distCombo.addItem("Gaussian Distribution");
	distCombo.addItem("Uniform Distribution");
	if (fd.isProbabilistic() && (fd.getDefaultValue() != null) ) {
	    ProbDist tmpPD = (ProbDist)fd.getDefaultValue();
	    if (tmpPD instanceof GaussianDist) {
		distCombo.setSelectedItem("Gaussian Distribution");
	    } else if (tmpPD instanceof UniformDist) {
		distCombo.setSelectedItem("Uniform Distribution");
	    }
	}
	gbl.setConstraints(distCombo, gbc);
	probPanel.add(distCombo);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	JButton b = new JButton("Create");
	b.setActionCommand("CREATE");
	gbl.setConstraints(b, gbc);
	probPanel.add(b);

	if (viewing) {
	    b.setEnabled(false);
	}
	
	// When a choice is made, replace the panel with one editing the
	// distribution.
	b.addActionListener(this);

	if (fd.isProbabilistic()) {
	    ContinuousFeatureDescriptor cfd = (ContinuousFeatureDescriptor)fd;
	    defPD = (ProbDist)((ProbDist)cfd.getDefaultValue()).clone();
	    Component c = defPD.getEditViewPanel(!viewing);
	    subProbPanel.add(c);
	}

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(subProbPanel, gbc);
	probPanel.add(subProbPanel);

    }

    public void actionPerformed(ActionEvent ae)
    {
	if (ae.getActionCommand().equals("CREATE")) {
	    String toCreate = (String)distCombo.getSelectedItem();
	    if (toCreate.equals("Gaussian Distribution")) {
		defPD = new GaussianDist();
	    } else if (toCreate.equals("Uniform Distribution")) {
		try {
		    defPD = new UniformDist( new Double( lowerBoundTF.getText() ).doubleValue(),
					     new Double( upperBoundTF.getText() ).doubleValue() );
		} catch (NumberFormatException nfe) {
		    // Shouldn't happen, but might if the human is screwed up.
		    defPD = new UniformDist();
		}
	    }
	    // Replace the existing subProbPanel
	    Component c = defPD.getEditViewPanel(true);
	    subProbPanel.removeAll();
	    subProbPanel.add(c);
	    probDetPanel.validate();
	    ((CardLayout)probDetPanel.getLayout()).show(probDetPanel, "Prob");
	} // end if
	super.actionPerformed(ae);
    }
    
    public void createDetPanel()
    {
	JLabel l;
	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();

	detPanel = new JPanel(gbl);
      
	gbc.fill = GridBagConstraints.NONE;
	gbc.gridwidth = GridBagConstraints.RELATIVE;
	l = new JLabel("Default Value:");
	gbl.setConstraints(l, gbc);
	detPanel.add(l);

	gbc.fill = GridBagConstraints.HORIZONTAL;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	String cdefVal = new String("");
	if (! fd.isProbabilistic()) {
	    cdefVal = nf.format( (Double)fd.getDefaultValue() );
	}
	defValTF = new JTextField( cdefVal, 10 );
	if (viewing) {
	    defValTF.setEditable(false);
	}
	gbl.setConstraints(defValTF, gbc);
	detPanel.add(defValTF);
    }
    
    public void canSaveValues(Vector reasons)
    {
	super.canSaveValues(reasons);
	// Two cases:
	if ( det.isSelected() ) {
	    // 1] Deterministic: See if the number parses correctly.
	    try {
		Number toSave = nf.parse( defValTF.getText() );
		// Plus, see if it's a valid value.
		if (! fd.isValidValue( new Double(toSave.doubleValue()) )) {
		    reasons.addElement("Default value \"" +
				       toSave.intValue() +
				       "\" is not a valid value [outside bounds?].");
		}
	    } catch (ParseException pe) {
		reasons.addElement("Cannot parse the default value \"" +
				   defValTF.getText() +
				   "\" as a number.");
	    } // end try-catch
	} else {
	    // 2] Probabilistic: Ask the ProbDist.
	    // We assume that the ProbDist can save it's values...
	    // Unless there's no in-progress-editing ProbDist object, in
	    // which case, we complain.
	    if (defPD == null) {
		reasons.addElement("No default probability distribution has been created.");
	    }
	} // end if-else
    }

    public String getFDTypeName()
    {
	return new String("Continuous Feature Descriptor");
    }


    public void saveValues()
	throws CannotSaveValuesException
    {
	if (viewing) { return; }
	super.saveValues();
	if (det.isSelected()) {
	    // Set the default value
	    try {
		// Lazy-create a number-formatter
		if (nf == null) {
		    nf = new DecimalFormat();
		}
		ContinuousFeatureDescriptor cfd =
		    (ContinuousFeatureDescriptor)fd;
		cfd.setDefaultValue( new Double(nf.parse(defValTF.getText()).
						doubleValue()) );
	    } catch (ParseException pe) {
		throw new CannotSaveValuesException(pe.getMessage());
	    } // end try-catch
	} else {
	    // Set the default dist
	    try {
		defPD.saveValues();
		fd.defaultValue = defPD;
	    } catch (NotBeingEditedException nbee) {
		// Won't happen as we're editing at the moment.
		throw new InternalError(nbee.getMessage());
	    } catch (CannotSaveValuesException csve) {
		// Shouldn't happen
		throw new InternalError(csve.getMessage());
	    } // end try-catch
	}
    }

}
