package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.*;
import java.text.*;
import java.util.Vector;
import NTLSystem.NTLException;

/**
   An integral feature-descriptor contains meta information that pertains
   to all instantiations of a particular integral feature.
   <P>
   $Id: IntegralFeatureDescriptor.java,v 1.16 1998/10/16 01:57:10 jsled Exp $
   <BR>Last editor: $Author: jsled $

   @author     Tajh Taylor [tajh@icsi.berkeley.edu]
   @version    $Revision: 1.16 $
 **/

public class IntegralFeatureDescriptor
extends BoundedFeatureDescriptor
{
  /**
   * serialization Stream Unique Identifier.  Recompute [with
   * serialver] when this class changes.
   **/
  static final long serialVersionUID = 1234348411811590251L;

  /* ATTRIBUTES */

   // public instance methods

   // constructors

   /** No-argument constructor for serialization. */
   public IntegralFeatureDescriptor()
   {
     super("", new Integer(0), new Integer(Integer.MAX_VALUE));
     try {
       setDefaultValue(new Integer(0));
     } catch (NTLException ntle) {
       // Shouldn't happen
       throw new InternalError(ntle.getMessage());
     } // end try-catch
   }

   /** Simple determinate constructor; assumes default bounds and
       default default value.
       @param     newName the new name for this feature */
   public IntegralFeatureDescriptor(String newName)
   {
      super(newName, new Integer(0), new Integer(Integer.MAX_VALUE));
      try {
	setDefaultValue( new Integer(0) );
      } catch (NTLException ntle) {
	// Shouldn't happen...
	throw new InternalError(ntle.getMessage());
      } // end try-catch
   }

   /** Detailed determinate constructor.
       Since a default value is given, this feature is assumed to
       be determinate.
       @param     newName       the new name for this feature
       @param     newLowerBound the new lower bound for the feature value range 
       @param     newUpperBound the new upper bound for the feature value range
       @param     newDefault    the new default value for this feature
       @exception NTLException  if setDefaultValue() throws one. */
   public IntegralFeatureDescriptor(String newName,
				    int newLowerBound,
				    int newUpperBound,
				    int newDefault)
      throws NTLException
   {
     super(newName, new Integer(newLowerBound), new Integer(newUpperBound));
     setDefaultValue(new Integer(newDefault));
   }

   /** Simple probabilistic constructor; assumes default bounds.
       @param newName        the new name for this feature
       @param newDefaultDist the new default distribution */
   public IntegralFeatureDescriptor(String newName, ProbDist newDefaultDist)
   {
      super(newName,
	    new Integer(0), new Integer(Integer.MAX_VALUE),
	    newDefaultDist);
   }

   /** Detailed probabilistic constructor.
       Since a default distribution is given, the feature is
       assumed to be probabilistic.
       @param newName        the new name for this feature
       @param newLowerBound  the new lower bound for the feature value range
       @param newUpperBound  the new upper bound for the feature value range
       @param newDefaultDist the new default distribution */
   public IntegralFeatureDescriptor(String newName,
				    int newLowerBound,
				    int newUpperBound,
				    ProbDist newDefaultDist)
   {
      super(newName,
	    new Integer(newLowerBound), new Integer(newUpperBound),
	    newDefaultDist);
   }

   // implementation of abstract FeatureDescriptor methods

   /** Report on whether or not this feature has ordered values;
       always return true.
       @return always returns true */
   public boolean isOrdered()
   {
      return true;
   }

  /** Test whether the argument object contains a valid value for
       this particular feature.  Note that implementations should 
       report false for ProbDist object if the feature is determinate,
       and for definite value object if the feature is probabilistic.
       @param o any object to be tested.
       @return boolean flag indicating validity. */
   public boolean isValidValue(Object o)
   {
      Integer i;
      ProbDist pd;
      
      // check and see if the feature is probabilistic first
      if (isProbabilistic())
      {
	 // make sure the distribution is probabilistic
	 if (o instanceof ProbDist)
	 {
	    pd = (ProbDist)o;
	    if (pd.isContinuous())
	       return true;
	    else
	       return false;
	 }
	 else
	    return false;
      }
      else // it's a determinate feature
      {
	 // make sure it's not a ProbDist
	 if (o instanceof ProbDist)
	    return false;
	 
	 // make sure it IS an int wrapped in an Integer
	 if (o instanceof Integer)
	 {
	    i = (Integer)o;

	    // check the bounds
	    if ( (i.intValue() >= getLowerBound().intValue()) &&
		 (i.intValue() <= getUpperBound().intValue()) ){
	       return true;
	    } else {
	       return false;
	    }
	 } else {
	    return false;
	 }
      }
   }

   // class-specific methods

   /** Set the default value for this feature.
       Makes the feature determinate.
       @param     newDefault the new default value for this feature
       @exception NTLException if the default value is outside this
                  descriptor's bounds. */
  protected void setDefaultValue(Integer newDefault)
       throws NTLException
  {
    if ( (newDefault.doubleValue() >= getLowerBound().doubleValue()) &&
	 (newDefault.doubleValue() <= getUpperBound().doubleValue()) ) {
      defaultValue = newDefault;
      setProbabilistic(false);
    } else {
      throw new NTLException("Default value must be within bounds.");
    }
  }

  /**
   * @see FeatureDescriptor
   **/
  public Object valueFromText(String textVal)
       throws IllegalArgumentException
  {
    Integer toRet = null;
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setParseIntegerOnly(true);
    try {
      toRet = new Integer((nf.parse(textVal)).intValue());
    } catch (ParseException pe) {
      throw new IllegalArgumentException(pe.getMessage());
    }
    if (toRet == null) {
      throw new IllegalArgumentException("No number parsed from " + textVal);
    }
    return toRet;
  }

  /**
   * @see FeatureDescriptor
   **/
  public String valueAsText(Object val)
  {
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setParseIntegerOnly(true);
    return nf.format( (Integer)val );
  }
   
   /** Test for equality with another object.
       @param o object to be tested
       @return boolean flag indicating equality */
   public boolean equals(Object o)
   {
     if (!super.equals(o)) { return false; }
     IntegralFeatureDescriptor ifd = (IntegralFeatureDescriptor)o;
     if (! getDefaultValue().equals(ifd.getDefaultValue()) ) {
       return false;
     }

     return true;
   }

   /** Read in the descriptor from human-readable text.
       @param input input source */
   public void readExternal(ObjectInput input)
      throws IOException
   {
      //TODO requires a generic word parser class
   }

   /** Write the descriptor out in human-readable text form.
       Completely overrides the parent method.
       @param output output sink
       @exception IOException if there is an IOException when writing
       the data.
       **/
   public void writeExternal(ObjectOutput output)
       throws IOException
   {
      // write a header
      output.writeChars("IntegralFeatureDescriptor\n");
   }

  public String getDisplayableName()
  {
    return new String("Integral Feature Descriptor");
  }

  protected JPanel getEditViewPanel(boolean edit)
  {
    viewer = new IFDViewer(this, edit);
    return viewer;
  }

}

class IFDViewer
    extends BFDViewer
{

    /**
     * The JTextField used for default value editing.
     **/
    JTextField defValTF;

    /**
     * The new ProbDist object [when editing].
     **/
    ProbDist defPD;
    /**
     * The ComboBox to use for selecting which type of Distribution to
     * create.
     **/
    JComboBox distCombo;
    /**
     * The sub-panel actually containing the ProbabilityDistribution editor.
     **/
    JPanel subProbPanel;

    public IFDViewer(FeatureDescriptor this_fd, boolean edit)
    {
	super(this_fd, edit);
    }

    public void createProbPanel()
    {
	// Create the ProbDist-creation panel.
	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	probPanel = new JPanel(gbl);
	subProbPanel = new JPanel();

	gbc.fill = GridBagConstraints.NONE;
	gbc.gridwidth = GridBagConstraints.RELATIVE;

	// Present the option to create a distribution.
	distCombo = new JComboBox();
	distCombo.setEditable(false);
	if (viewing) {
	    distCombo.setEnabled(false);
	}
	distCombo.addItem("Gaussian Distribution");
	distCombo.addItem("Uniform Distribution");
	if (fd.isProbabilistic() && (fd.getDefaultValue() != null) ) {
	    ProbDist tmpPD = (ProbDist)fd.getDefaultValue();
	    if (tmpPD instanceof GaussianDist) {
		distCombo.setSelectedItem("Gaussian Distribution");
	    } else if (tmpPD instanceof UniformDist) {
		distCombo.setSelectedItem("Uniform Distribution");
	    }
	}
	gbl.setConstraints(distCombo, gbc);
	probPanel.add(distCombo);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	JButton b = new JButton("Create");
	b.setActionCommand("CREATE");
	gbl.setConstraints(b, gbc);
	probPanel.add(b);
	if (viewing) {
	    b.setEnabled(false);
	}
	
	// When a choice is made, replace the panel with one editing the
	// distribution.
	b.addActionListener(this);
      
	if (fd.isProbabilistic()) {
	    IntegralFeatureDescriptor ifd = (IntegralFeatureDescriptor)fd;
	    defPD = (ProbDist)((ProbDist)ifd.getDefaultValue()).clone();
	    Component c = defPD.getEditViewPanel(!viewing);
	    subProbPanel.add(c);
	}
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(subProbPanel, gbc);
	probPanel.add(subProbPanel);
    }

    public void actionPerformed(ActionEvent ae)
    {
	if (ae.getActionCommand().equals("CREATE")) {
	    String toCreate = (String)distCombo.getSelectedItem();
	    if (toCreate.equals("Gaussian Distribution")) {
		defPD = new GaussianDist();
	    } else if (toCreate.equals("Uniform Distribution")) {
		try {
		    defPD = new UniformDist( new Double( lowerBoundTF.getText() ).doubleValue(),
					     new Double( upperBoundTF.getText() ).doubleValue());
		} catch (NumberFormatException nfe) {
		    // Shouldn't happen, but might if the human is messed up.
		    defPD = new UniformDist();
		}
	    }
	    // Replace the existing subProbPanel
	    Component c = defPD.getEditViewPanel(true);
	    subProbPanel.removeAll();
	    subProbPanel.add(c);
	    probDetPanel.validate();
	    ((CardLayout)probDetPanel.getLayout()).show(probDetPanel, "Prob");
	} // end if
	super.actionPerformed(ae);
    }
    
    public void createDetPanel()
    {
	JLabel l;
	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();

	detPanel = new JPanel(gbl);
      
	gbc.fill = GridBagConstraints.NONE;
	gbc.gridwidth = GridBagConstraints.RELATIVE;
	l = new JLabel("Default Value:");
	gbl.setConstraints(l, gbc);
	detPanel.add(l);

	gbc.fill = GridBagConstraints.HORIZONTAL;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	String idefVal = new String("");
	if (! fd.isProbabilistic()) {
	    idefVal = nf.format( (Integer)fd.getDefaultValue() );
	}
	defValTF = new JTextField( idefVal, 10 );
	if (viewing) {
	    defValTF.setEditable(false);
	}
	gbl.setConstraints(defValTF, gbc);
	detPanel.add(defValTF);
    }
    
    public void canSaveValues(Vector reasons)
    {
	super.canSaveValues(reasons);
	// Two cases:
	if ( det.isSelected() ) {
	    // 1] Deterministic: See if the number parses correctly.
	    try {
		Number toSave = nf.parse( defValTF.getText() );
		// Plus, see if it's a valid value.
		if (! fd.isValidValue( new Integer(toSave.intValue()) )) {
		    reasons.addElement("Default value \"" +
				       toSave.intValue() +
				       "\" is not a valid value [outside bounds?].");
		}
	    } catch (ParseException pe) {
		reasons.addElement("Cannot parse the default value \"" +
				   defValTF.getText() +
				   "\" as a number.");
	    } // end try-catch
	} else {
	    // 2] Probabilistic: Ask the ProbDist.
	    // We assume that the ProbDist can save it's values...
	    // Unless there's no in-progress-editing ProbDist object, in
	    // which case, we complain.
	    if (defPD == null) {
		reasons.addElement("No default probability distribution has been created.");
	    }
	} // end if-else
    }

    public String getFDTypeName()
    {
	return new String("Integral Feature Descriptor");
    }


    public void saveValues()
	throws CannotSaveValuesException
    {
	if (viewing) { return; }
	super.saveValues();
	if (det.isSelected()) {
	    // Set the default value
	    try {
		if (nf == null) {
		    nf = new DecimalFormat();
		}
		((IntegralFeatureDescriptor)fd).
		    setDefaultValue(new Integer((nf.parse(defValTF.getText())).intValue()));
	    } catch (ParseException pe) {
		throw new CannotSaveValuesException(pe.getMessage());
	    } catch (NTLException ntle) {
		// Shouldn't happen, as saveValues() shouldn't have been
		// called if canSaveValues() returned non-null...
		throw new CannotSaveValuesException(ntle.getMessage());
	    } // end try-catch
	} else {
	    // Set the default dis
	    try {
		defPD.saveValues();
		fd.defaultValue = defPD;
	    } catch (NotBeingEditedException nbee) {
		// Won't happen as we're editing at the moment.
		throw new InternalError(nbee.getMessage());
	    } catch (CannotSaveValuesException csve) {
		// Shouldn't happen
		throw new InternalError(csve.getMessage());
	    } // end try-catch
	}
    }

}
