package NTLSystem.support;

import java.lang.reflect.*;
import java.beans.*;

/** 
 * A descriptor for an editable property.  Classes which implement the
 * PropertyCollection interface use these objects to return a
 * description of an editable property, which allows the
 * PropertyCollectionSheet to edit the data.
 * <P>
 * $Id: EditablePropertyDescriptor.java,v 1.5 1998/06/24 23:05:05 jsled Exp $<BR>
 * Last Editor: $Author: jsled $
 * 
 * @author    Josh Sled [jsled@icsi.berkeley.edu]
 * @see       PropertyCollectionSheet
 * @see       PropertyCollection
 * @see       PropertyDescriptor
 * @version   $Revision: 1.5 $
 * 
 **/
public class EditablePropertyDescriptor
{

  /* ATTRIBUTES */

  /** The name of the property **/
  private String name;
  /** The type of the Property **/
  private Class type;
  /** The getter method for the Property **/
  private Method getter;
  /** The setter method for the property **/
  private Method setter;
  /**
   * An index for the property's getter/setter method which indexes
   * the property. 
   **/
  private Object index;
  /**
   * The object on which the Methods should be called.
   **/
  private Object target;


  /* CONSTRUCTOR */
  
  /**
   * A constructor for EditablePropertyDescriptors.
   *
   * @param     name The name of the property.
   * @param     type This property's type.
   * @param     getter The Method to use for getting the value.
   * @param     setter The Method to use to set the value.
   * @param     index An index for the getter/setter method which
   *            indexes the property. If null, the getter and setter
   *            values are used as-is.  If non-null, used to index the
   *            getter/setter methods.
   * @param     target The target object for this property.  This is
   *            the object on which the get and set methods should be
   *            called. If null, the object which is returning the
   *            EditablePropertyDescriptors should be used.
   **/
  public EditablePropertyDescriptor(String name,
				    Class type,
				    Method getter,
				    Method setter,
				    Object index,
				    Object target)
  {
    this.name = name;
    this.type = type;
    this.getter = getter;
    this.setter = setter;
    this.index = index;
    this.target = target;
  }

  /**
   * A convenience constructor for EditablePropertyDescriptors whos
   * properties don't require an index argument to their
   * getter/setter methods or a specified target object.
   **/
  public EditablePropertyDescriptor(String name,
				    Class type,
				    Method getter,
				    Method setter)
  {
    this(name, type, getter, setter, null, null);
  }

  /**
   * A convenience constructor for EditablePropertyDescriptors which
   * are using the PropertyCollection implementer as the target for
   * getter/setter invocation.
   **/
  public EditablePropertyDescriptor(String name,
				    Class type,
				    Method getter,
				    Method setter,
				    Object index)
  {
    this(name, type, getter, setter, index, null);
  }


  /* METHODS */

  /**
   * @return     Returns the name of the Property
   **/
  public String getName()
  {
    return new String(name);
  }


  /**
   * Gets the type of this property.
   *
   * @return     The type of this property.
   **/ 
  public Class getType()
  {
    return type;
  }

  /**
   * @return     The getter Method for the property.
   **/
  public Method getReadMethod()
  {
    return getter;
  }


  /**
   * @return    The setter Method for the property.
   **/
  public Method getWriteMethod()
  {
    return setter;
  }

  /**
   * Gets the index object for this property. If this index is
   * non-null, it should be fed as the first argument to the
   * getter/setter method to get/set the correct property.  If null,
   * the getter/setter method refers to the property directly, without
   * needing to be indexed.
   *
   * @return    The index object for this property.
   **/
  public Object getIndexer()
  {
    return index;
  }

  /**
   * Returns the target object on which to invoke the getter/setter
   * methods.  If null, the object which implements PropertyCollection
   * should be used as the target for invocation.
   *
   * @return    The target for getter/setter method invocation.
   **/
  public Object getTarget()
  {
    return target;
  }
}

