package NTLSystem.support;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * A extension of Panel which actually displays and allows editing of
 * the values in the PropertyCollection.  This code was stolen from
 * the BDK's BeanBox and heavily modified for our needs.
 * <P>
 * The modifications are as such:
 * <OL>
 * <LI>Removed ability to embed itself within a ScrollPane.
 * <LI>Switched GUI over to Swing 1.0.2;  Implements Scrollable to be
 *     put in a JScrollPane.
 * <LI>Custom editors are embedded within the Panel itself.  Custom
 *     editors should be small, but have the ability to pop up a large
 *     dialog with full editing ability.
 *     <TT>PropertyEditor.isPaintable()</TT> is not supported, as the
 *     editor should use a custom editor.
 * <LI>Bound and vetoable properties are not supported.
 * </OL>
 * <P>
 * $Id: PropertyCollectionSheetPanel.java,v 1.10 1998/08/19 02:50:24 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author Josh Sled [jsled@icsi.berkeley.edu]
 * @see PropertyCollection
 * @version $Revision: 1.10 $
 **/
public class PropertyCollectionSheetPanel
extends JPanel
implements PropertyChangeListener
{

  public PropertyCollectionSheetPanel() {
    super();
    /* We want 2 columns and as many rows as necessary... as such:
     *  | label_0 | view_0 |
     *  | label_1 | view_1 |
     *  |   ...   |  ...   |
     *  | label_n | view_n |
     *
     * However, we don't want the labels to have the same width as the
     * views... so we use a GridBagLayout.
     */
    saveAsEdited = false;
    gbl = new GridBagLayout();
    setLayout(gbl);
  }

  public PropertyCollectionSheetPanel(PropertyCollection pcObj)
  {
    this( );
    setTarget(pcObj);
  }

  public synchronized void setTarget(PropertyCollection pcObj) {

    removeAll();

    // We make the panel invisivle during the reconfiguration
    // to try to reduce screen flicker.

    // As a workaround for #4056424, we avoid maling the panel
    // invisible first time though, during startup.
    if (pcObj != null) {
      setVisible(false);
    } else {
	return;
    }

    // Save the target object being edited.  This object implements
    // the PropertyCollection interface.
    target = pcObj;

    properties = target.getEditableProperties();
    
    editors = new PropertyEditor[properties.length];
    values = new Object[properties.length];
    views = new Component[properties.length];
    labels = new JLabel[properties.length];

    // When adding the labels and views at the end of the for loop.
    GridBagConstraints c = new GridBagConstraints();
    c.fill = GridBagConstraints.BOTH;

    for (int i = 0; i < properties.length; i++) {

      String name = properties[i].getName();
      Class type = properties[i].getType();
      Method getter = properties[i].getReadMethod();
      Method setter = properties[i].getWriteMethod();

      if ( getter == null || setter == null ) {
	warning("No getter/setter method specified for property \"" +
		name + "\" in object \"" +
		target.getLabel() + "\".");
	continue;
      }

      Component view = null;
      

      try {
	Object index = properties[i].getIndexer();
	Vector args = new Vector(1);
	if (index != null) {
	  args.addElement(index);
	}
	Object argsA[] = new Object[args.size()];
	args.copyInto(argsA);
	Object invokee = properties[i].getTarget();
	if (invokee == null) { invokee = target; }
	Object value = getter.invoke(invokee, argsA);
	values[i] = value;

	PropertyEditor editor = null;
	editor = PropertyEditorManager.findEditor(type);
	editors[i] = editor;

	// If we can't edit this component, skip it.
	if (editor == null) {
	  // If it's a user-defined property we give a warning.
	  String getterClass = properties[i].getReadMethod().getDeclaringClass().getName();
	  if (getterClass.indexOf("java.") != 0) {
	    System.err.println("Warning: Can't find public property editor for property \""
+ name + "\".  Skipping.");
	  }
	  continue;
	}

	// Don't try to set null values:
	if (value == null) {
	  // If it's a user-defined property we give a warning.
	  String getterClass = properties[i].getReadMethod().getDeclaringClass().getName();
	  if (getterClass.indexOf("java.") != 0) {
	    System.err.println("Warning: Property \"" + name 
			       + "\" has null initial value.  Skipping.");	
	  }
	  continue;
	}

	editor.setValue(value);
	editor.addPropertyChangeListener(this);

	// Now figure out how to display it...
	if (editor.supportsCustomEditor()) {
	  view = editor.getCustomEditor();
	} else if (editor.getTags() != null) {
	  view = new PropertySelector(editor);
	} else if (editor.getAsText() != null) {
	  view = new PropertyText(editor);
	} else {
	  System.err.println("Warning: Property \"" + name 
			     + "\" has non-displayabale editor.  Skipping.");
	  continue;
	}

      } catch (InvocationTargetException ex) {
	System.err.println("Skipping property " + name + " ; exception on target: " + ex.getTargetException());
	ex.getTargetException().printStackTrace();
	continue;
      } catch (Exception ex) {
	System.err.println("Skipping property " + name + " ; exception: " + ex);
	ex.printStackTrace();
	continue;
      }

      // Break the name up into lines
      Vector lines = new Vector();
      int lastIdx = 0;
      int idx = 0;
      while ( (idx = name.indexOf('\n', lastIdx)) != -1 ) {
	  lines.addElement( name.substring(lastIdx, idx) );
	  lastIdx = idx + 1;
      }
      // Get the last string.
      lines.addElement( name.substring(lastIdx, name.length()) );
      // Create a label for the first line.  The rest are used as
      // ToolTip text.
      labels[i] = new JLabel((String)lines.elementAt(0), Label.RIGHT);
      StringBuffer ttText = new StringBuffer( name.length() -
					      ((String)lines.elementAt(0)).length() );
      for (int j=1; j<lines.size(); j++) {
	  ttText.append( (String)lines.elementAt(j) );
	  if (j != lines.size()-1) {
	      ttText.append( ", " );
	  }
      } // end for over j
      labels[i].setToolTipText( ttText.toString() );
      // Pad the label a bit [3-pixel border on all sides].
      labels[i].setBorder( BorderFactory.createEmptyBorder(3, 3, 3, 3) );
      c.fill = GridBagConstraints.NONE;
      c.gridwidth = GridBagConstraints.RELATIVE;
      c.anchor = GridBagConstraints.EAST;
      gbl.setConstraints(labels[i], c);
      add(labels[i]);
      
      c.fill = GridBagConstraints.HORIZONTAL;
      c.gridwidth = GridBagConstraints.REMAINDER;
      c.anchor = GridBagConstraints.WEST;
      views[i] = view;
      gbl.setConstraints(views[i], c);
      add(views[i]);
      
    }

    doLayout();
    setVisible(true);
	
  }

    /**
     * Sets the state of the panel with regards to saving values as
     * their edited.  If set to true, values will be saved back into
     * the being-edited object as they are modified.  If false,
     * 'saveValues' must be called to save the being-edited values.
     *
     * @param     saveAsEdited If true, save values as they are
     *            edited, if false, save only when saveValues() is
     *            called.
     **/
    public void setSaveAsEdited( boolean saveAsEdited )
    {
	this.saveAsEdited = saveAsEdited;
    }

  /**
   * Saves the values in the PropertyCollectionSheetPanel back to
   * their respective locations.  If one of the values is not in the
   * correct format to be saved an error dialog will be popped up, and
   * false will be returned.  If all the values are of the correct
   * format, true will be returned.
   *
   * @return    True if the properties were able to be saved, false
   *            otherwise.
   **/
  public synchronized boolean saveValues() {
    boolean toRet = true;

    // This goes through and updates values[] to all the "new" values
    // from each editor.
    for (int i = 0 ; i < editors.length; i++) {
      EditablePropertyDescriptor property = properties[i];
      Method setter = properties[i].getWriteMethod();
      Object value = editors[i].getValue();
      values[i] = value;
      try {
	Object index = property.getIndexer();
	Vector args = new Vector(2);
	if (index != null) {
	  args.addElement(index);
	}
	args.addElement(value);
	Object argsA[] = new Object[args.size()];
	args.copyInto(argsA);
	Object invokee = property.getTarget();
	if (invokee == null) { invokee = target; }
	setter.invoke(invokee, argsA);
	
      } catch (InvocationTargetException ex) {
	error("InvocationTargetException while updating " +
	      property.getName(), ex.getTargetException());
	toRet = false;
      } catch (Exception ex) {
	error("Unexpected exception while updating " +
	      property.getName(), ex);
	toRet = false;
      }
      if (views[i] != null) {
	views[i].repaint();
      }
    } // end for

    // Now re-read all the properties and update the editors
    // for any other properties that have changed.
    for (int i = 0; i < properties.length; i++) {
      Object o;
      try {
	Method getter = properties[i].getReadMethod();
	Object index = properties[i].getIndexer();
	Vector args = new Vector(1);
	if (index != null) {
	  args.addElement(index);
	}
	Object argsA[] = new Object[args.size()];
	args.copyInto(argsA);
	Object invokee = properties[i].getTarget();
	if (invokee == null) { invokee = target; }
	o = getter.invoke(invokee, argsA);
      } catch (Exception ex) {
	o = null;
      }
      if (o == values[i] || (o != null && o.equals(values[i]))) {
	// The property is equal to its old value.
	continue;
      } else {
	// The property is not equal to it's old value, but wasn't
	// changed above.  Therefore, there was a problem the user
	// needs to know about...
	warning("Property " + properties[i].getName() + " (type " +
		properties[i].getType().toString() + ") was not able " +
		" to be changed to value " + values[i].toString());
	toRet = false;
	values[i] = o;
      } 
      // Make sure we have an editor for this property...
      if (editors[i] == null) {
	continue;
      }
      // The property has changed!  Update the editor.
      editors[i].setValue(o);

      if (views[i] != null) {
	views[i].repaint();
      }

    }

    // Make sure the target bean gets repainted.
    if (Beans.isInstanceOf(target, Component.class)) {
      ((Component)(Beans.getInstanceOf(target, Component.class))).repaint();
    }

    return toRet;
  }

    /**
     * Returns the minimum size of this Component.  This is the
     * minimum width to hold all the field labels/views [bounded below
     * by minWidth, and the minimum height [bounded below by minHeight
     * and above by maxHeight].
     *
     * @return    The minimum Dimensions of this Component [Panel].
     **/
    public Dimension getMinimumSize()
    {
	// Get PCSP dimensions.
	Dimension dim = getPCSPSize();

	// Bound the width.
	if (dim.width < minWidth) {
	    dim.width = minWidth;
	}

	// Bound the height.
	if (dim.height > minHeight) {
	    dim.height = minHeight;
	}

	return dim;
    }

    /**
     * Returns the preferred size of this Component.  This is the
     * preferred width to hold all the field labels/views [bounded
     * below by minWidth], and the preferred height [bounded below by
     * minHeight]. 
     *
     * @return    The preferred Dimensions of this Component [Panel].
     **/
    public Dimension getPreferredSize()
    {
	// Get PCSP dimensions.
	Dimension dim = getPCSPSize();
	
	// Bound the width.
	if (dim.width < minWidth) {
	    dim.width = minWidth;
	}

	// Bound the height.
	if (dim.height < minHeight) {
	    dim.height = minHeight;
	}

	return dim;
    }

    /**
     * Goes through all labels[] and views[] and gets the total of
     * their sizes.  The max width and the total height are returned,
     * where 'width' is the maximum width of label+view, and
     * 'height' is the sum over all views of
     * max( label.height, view.height ) .
     * 
     * @return    Dimensions of the PCSP.
     **/
    protected Dimension getPCSPSize()
    {
	int width = 0;
	int height = 0;

	int tmpWidth, tmpHeight;
	Dimension label, view, emptyView = new Dimension(0, 0);
	if ( (views != null) && (labels != null) ) {
	    for (int i=0; i<views.length; i++) {
		
		tmpWidth = tmpHeight = 0;
		label = labels[i].getPreferredSize();
		view = emptyView; // to satisfy compiler warning.
		if (views[i] != null) {
		    view = views[i].getPreferredSize();
		}
		
		tmpWidth = label.width + view.width;
		tmpHeight = Math.max( label.height, view.height );
		
		width = Math.max( width, tmpWidth );
		height += tmpHeight;
	    } // end for over i
	}

	return new Dimension( width, height );
    }

  public void propertyChange(PropertyChangeEvent evt)
  {
    wasModified(evt);
  }

  public void wasModified(PropertyChangeEvent evt)
  {
    // Get the editor which is sourced the event.
    PropertyEditor editor = (PropertyEditor)evt.getSource();
    // Repaint the editor which was changed.
    for (int i=0; i<editors.length; i++) {
      if ( (editors[i] == editor) ) {
	  if (views[i] != null) {
	      views[i].repaint();
	  }

	  // If we're supposed to save as things are modified, then so
	  // do so.
	  if (saveAsEdited) {
	      EditablePropertyDescriptor property = properties[i];
	      Method setter = properties[i].getWriteMethod();
	      Object value = editors[i].getValue();
	      values[i] = value;
	      try {
		  Object index = property.getIndexer();
		  Vector args = new Vector(2);
		  if (index != null) {
		      args.addElement(index);
		  }
		  args.addElement(value);
		  Object argsA[] = new Object[args.size()];
		  args.copyInto(argsA);
		  Object invokee = property.getTarget();
		  if (invokee == null) { invokee = target; }
		  setter.invoke(invokee, argsA);
	
	      } catch (InvocationTargetException ex) {
		  error("InvocationTargetException while updating " +
			property.getName(), ex.getTargetException());
	      } catch (Exception ex) {
		  error("Unexpected exception while updating " +
			property.getName(), ex);
	      } // end try-catch
	  } // end if saveAsEdited
      } // end if editor matches
    } // end for
  } // end wasModified
    

  private void warning(String s) {
    JOptionPane.showMessageDialog(null, "Warning: " + s, "Warning...", JOptionPane.WARNING_MESSAGE);
  }

  //----------------------------------------------------------------------
  // Log an error.

  private void error(String message, Throwable th) {
    String mess = message + ":\n" + th;
    System.err.println(message);
    th.printStackTrace();
    // Popup an ErrorDialog with the given error message.
    JOptionPane.showMessageDialog(null, mess, "Error...", JOptionPane.ERROR_MESSAGE);

  }

  //----------------------------------------------------------------------
  private GridBagLayout gbl;

  private PropertyCollection target;
  private EditablePropertyDescriptor properties[];
  private PropertyEditor editors[];
  private Object values[];
  private Component views[];
  private JLabel labels[];
    private boolean saveAsEdited;

  private static int hPad = 3;
  private static int vPad = 3;

    private int minHeight = 200;
    private int maxHeight = 300;
    private int minWidth = 150;
}

