package NTLSystem.verblearn;

import java.io.*;
import com.sun.java.swing.*;
import com.sun.java.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Enumeration;
import java.util.Vector;
import NTLSystem.support.*;

/**
 * An internal frame for editing the values in a Dataset.  Given a
 * ScenarioSet, this frame will allow the user to manipulate the
 * different sets of Scenarios in a Dataset, as well as allow the user
 * to create a Dataset by specifying precentages, and randomly
 * selecting Scenarios from the ScenarioSet.  Changes are saved when
 * the window is closed, or the "Save Dataset" button is clicked
 * upon. 
 * <P>
 * $Id: DatasetEditor.java,v 1.4 1999/08/03 21:02:58 nchang Exp $<BR>
 * Last editor: $Author: nchang $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Dataset
 * @version   $Revision: 1.4 $
 **/
public class DatasetEditor
    extends JInternalFrame
{
  static final long serialVersionUID = -2983249527502760505L;
    /** The Dataset we're editing. **/
    protected Dataset dset;

    /** The ScenarioSet to draw the Scenarios from. **/
    protected ScenarioSet sset;

    /** A count of the Scenarios in the ScenarioSet. **/
    protected int scenCt;

    /** A JList for all the Scenarios in the ScenarioSet. **/
    private JList scenarioList;

    /** A JTextField containing the training percentage. **/
    private JTextField trainingPct;

    /** Our private copy of the training Scenarios.  We use this
     * because Dataset only returns Enumerations. **/
    private Vector training;

    /** A JList for all the training Scenarios. **/
    private JList trainingList;

    /** A JTextField containing the testing percentage. **/
    private JTextField testingPct;

    /** A JTextField containing the batch percentage. nc temp**/
    private JTextField batchPct;

    /** Our private copy of the testing Scenarios.  We use this
	because Dataset only returns Enumerations. **/
    private Vector testing;

    /** A JList for all the testing Scenarios. **/
    private JList testingList;

    /** A JCheckBox for the 'ensure disjoint' option which ensures
	that the training and testing sets are disjoint. **/
    private JCheckBox ensureDisjoint;

    /** The size of the padding to the left and right of the Lists.
	Double this value is the actual spacing between lists [as a
	borders of this size are around each list]. **/
    private static final int hPad = 4;

    /** How many rows to show in the lists. **/
    private static final int ROW_COUNT = 20;
    /** How wide the lists should be. **/
    private static final int LIST_WIDTH = 100;
    /** The "Add Selected" button text. **/
    private static final String ADD_SEL = "Add Selected";
    /** The "Add Selected" button tool tip text. **/
    private static final String ADD_SEL_TT = "Add selected Scenarios from Scenario list.";
    /** The "Remove Selected" button text. **/
    private static final String REM_SEL = "Remove Selected";
    /** The "Remove Selected" button toop tip text. **/
    private static final String REM_SEL_TT = "Remove selected Sceanrios from this list.";


    /**
     * Creates a new DatasetEditor based on the given Dataset and
     * ScenarioSet. 
     *
     * @param     dset The Dataset to edit.
     * @param     sset The ScenarioSet which contains this Dataset,
     *            and from which to draw Scenario names.
     **/
    public DatasetEditor(Dataset dset, ScenarioSet sset)
    {
	super("Dataset Editor - " + dset.getName(),
	      true, true, true, true);

	this.sset = sset;
	this.dset = dset;
	// Get the Scenario names from the Enumerations returned by
	// the Dataset.
	getScenarioNameVectors();

	JComponent jc = (JComponent)getContentPane();

	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();

	jc.setLayout( gbl );

	// Create the top row of labels.
	c.fill = GridBagConstraints.HORIZONTAL;
	c.anchor = GridBagConstraints.CENTER;
	c.ipadx = hPad;

	JLabel jl = new JLabel( "Scenario List", JLabel.CENTER );
	gbl.setConstraints( jl, c );
	jc.add( jl );

	c.gridwidth = GridBagConstraints.RELATIVE;
	jl = new JLabel( "Train", JLabel.CENTER );
	gbl.setConstraints( jl, c );
	jc.add( jl );

	c.gridwidth = GridBagConstraints.REMAINDER;
	jl = new JLabel( "Test", JLabel.CENTER );
	gbl.setConstraints( jl, c );
	jc.add( jl );

	// Create the second row of the scenario count/percentage
	// textfields
	c.gridwidth = 1;
	
	scenCt = sset.countScenarios();
	String addendum = (scenCt == 1) ? " Scenario" : " Scenarios";
	JLabel scenarioCount = new JLabel( scenCt + addendum,
					   JLabel.CENTER );
	gbl.setConstraints( scenarioCount, c );
	jc.add( scenarioCount );

	c.gridwidth = GridBagConstraints.RELATIVE;
	trainingPct = new JTextField( 4 );
	gbl.setConstraints( trainingPct, c );
	jc.add( trainingPct );

	c.gridwidth = GridBagConstraints.REMAINDER;
	testingPct = new JTextField( 4 );
	gbl.setConstraints( testingPct, c );
	jc.add( testingPct );

	// Fill in these percentages
	updatePercentages();
	
	// Create the third row with the ever-useful JLists.
	c.gridwidth = 1;

	scenarioList = new JList( sset.scenarios );
	scenarioList.setVisibleRowCount( ROW_COUNT );
	scenarioList.setFixedCellWidth( LIST_WIDTH );
	scenarioList.setSelectionMode( ListSelectionModel.
				       MULTIPLE_INTERVAL_SELECTION);
	JScrollPane jsp =
	    new JScrollPane( scenarioList,
			     JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			     JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
	jsp.setBorder( BorderFactory.
		       createCompoundBorder(BorderFactory.
					    createEmptyBorder(0, hPad,
							      0, hPad),
					    jsp.getBorder() ) );
	gbl.setConstraints( jsp, c );
	jc.add( jsp );

	c.gridwidth = GridBagConstraints.RELATIVE;
	trainingList = new JList( training );
	trainingList.setVisibleRowCount( ROW_COUNT );
	trainingList.setFixedCellWidth( LIST_WIDTH );
	trainingList.setSelectionMode( ListSelectionModel.
				       MULTIPLE_INTERVAL_SELECTION);
	jsp =
	    new JScrollPane( trainingList,
			     JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			     JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
	jsp.setBorder( BorderFactory.
		       createCompoundBorder(BorderFactory.
					    createEmptyBorder(0, hPad,
							      0, hPad),
					    jsp.getBorder() ) );
	gbl.setConstraints( jsp, c );
	jc.add( jsp );

	c.gridwidth = GridBagConstraints.REMAINDER;
	testingList = new JList( testing );
	testingList.setVisibleRowCount( ROW_COUNT );
	testingList.setFixedCellWidth( LIST_WIDTH );
	testingList.setSelectionMode( ListSelectionModel.
				      MULTIPLE_INTERVAL_SELECTION);
	jsp =
	    new JScrollPane( testingList,
			     JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			     JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
	/*
	jsp.setBorder( BorderFactory.
		       createCompoundBorder(BorderFactory.
					    createEmptyBorder(0, hPad,
							      0, hPad),
					    jsp.getBorder() ) );
	*/
	
	gbl.setConstraints( jsp, c );
	jc.add( jsp );

	// Add second-to-last row of 'add selected' buttons.
	c.gridwidth = 1;

	// Add a blank label spacer for the "Scenario List" column.
	jl = new JLabel( "" );
	gbl.setConstraints( jl, c );
	jc.add( jl );

	c.gridwidth = GridBagConstraints.RELATIVE;
	JButton jb = new JButton( ADD_SEL );
	jb.setToolTipText( ADD_SEL_TT );
	jb.addActionListener( new AddButtonActionListener( training,
							   trainingList ) );
	gbl.setConstraints( jb, c );
	jc.add( jb );

	c.gridwidth = GridBagConstraints.REMAINDER;
	jb = new JButton( ADD_SEL );
	jb.setToolTipText( ADD_SEL_TT );
	jb.addActionListener( new AddButtonActionListener( testing,
							   testingList ) );
	gbl.setConstraints( jb, c );
	jc.add( jb );

	// Create the last row of 'remove selected' buttons.
	c.gridwidth = 1;
	
	// Add a blank spacer for the ScenarioList.
	jl = new JLabel( "" );
	gbl.setConstraints( jl, c );
	jc.add( jl );

	c.gridwidth = GridBagConstraints.RELATIVE;
	jb = new JButton( REM_SEL );
	jb.setToolTipText( REM_SEL_TT );
	jb.addActionListener( new RemoveButtonActionListener( training,
							      trainingList ) );
	gbl.setConstraints( jb, c );
	jc.add( jb );

	c.gridwidth = GridBagConstraints.REMAINDER;
	jb = new JButton( REM_SEL );
	jb.setToolTipText( REM_SEL_TT );
	jb.addActionListener( new RemoveButtonActionListener( testing,
							      testingList ) );
	gbl.setConstraints( jb, c );
	jc.add( jb );

	// Add a very last row with the 'save'[update] and 'generate'
	// buttons. nc + 'save to file'
	GridBagLayout gbl2 = new GridBagLayout();
	GridBagConstraints c2 = new GridBagConstraints();
	JPanel jp = new JPanel( gbl2 );

	c2.gridwidth = GridBagConstraints.REMAINDER;
	c2.fill = GridBagConstraints.HORIZONTAL;
	c2.anchor = GridBagConstraints.CENTER;
	ensureDisjoint = new JCheckBox( "Ensure disjoint sets", true );
	gbl2.setConstraints( ensureDisjoint, c2 );
	jp.add( ensureDisjoint );

	// next row: generate/save buttons
	c2.gridwidth = 1;
	//	c2.fill = GridBagConstraints.NONE;

	jb = new JButton( "Generate" );
	jb.setToolTipText( "Generate Dataset entries based on percentages." );
	jb.addActionListener( new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		// Get the percentage from the textfields [make sure
		//   to strip '%' if existant].
		String trainPctStr = trainingPct.getText();
		String testPctStr = testingPct.getText();
		int trainSymIdx = trainPctStr.lastIndexOf("%");
		int testSymIdx = testPctStr.lastIndexOf("%");
		if (trainSymIdx != -1) {
		    trainPctStr = trainPctStr.substring(0,
							trainSymIdx-1);
		}
		if (testSymIdx != -1) {
		    testPctStr = testPctStr.substring(0,
						      testSymIdx-1);
		}
		double trainWeight = 0;
		double testWeight = 0;
		try {
		    trainWeight =
			(new Double(trainPctStr)).doubleValue();
		    System.out.println("Train weight: " + trainWeight);
		    testWeight =
			(new Double(testPctStr)).doubleValue();
		    System.out.println("Test weight: " + testWeight);
		} catch (NumberFormatException nfe) {
		    JOptionPane.showMessageDialog(DatasetEditor.this,
						  "Illegal percentage value.",
						  nfe.getMessage(),
						  JOptionPane.ERROR_MESSAGE);
		    return; // error.
		}
		// Convert percentages to doubles
		trainWeight /= 100;
		testWeight /= 100;

		DatasetEditor.this.dset.
		    setsFromScenarios( DatasetEditor.this.sset.
				       getAllScenarios(),
				       trainWeight,
				       testWeight,
				       true,
				       ensureDisjoint.isSelected() );
		// Update GUI
		getScenarioNameVectors();
		trainingList.setListData( training );
		testingList.setListData( testing );
		trainingList.validate();
		testingList.validate();
		updatePercentages();
	    }
	} );
	gbl2.setConstraints( jb, c2 );
	jp.add( jb );

	// not sure why, but this layout only shows 2 buttons; leave this one out for now!
	c2.gridwidth = GridBagConstraints.RELATIVE;

	jb = new JButton( "Save" );
	jb.setToolTipText( "Update Dataset with new values." );
	jb.addActionListener( new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		updateDataset();
	    }
	} );
	gbl2.setConstraints( jb, c2 );
	jp.add( jb );
	
	// nc added
	c2.gridwidth = GridBagConstraints.REMAINDER;

	jb = new JButton( "Save As" );
	jb.setToolTipText( "Save Dataset to file and add to this ScenarioSet." );
	jb.addActionListener( new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
	      saveDatasetToFile();
	    }
	} );
	gbl2.setConstraints( jb, c2 );
	jp.add( jb );
	

	c.gridwidth = GridBagConstraints.REMAINDER;
	c.fill = GridBagConstraints.NONE;
	c.anchor = GridBagConstraints.CENTER;
	gbl.setConstraints( jp, c );
	jc.add( jp );

	// Create a final row with batch saving option.
	c.gridwidth = 1;
	
	// Add a blank label spacer for the "Scenario List" column.
	jl = new JLabel( "" );
	gbl.setConstraints( jl, c );
	jc.add( jl );

	jl = new JLabel( "Batch size", JLabel.CENTER );
	gbl.setConstraints( jl, c );
	jc.add( jl );

	c.gridwidth = GridBagConstraints.RELATIVE;
	batchPct = new JTextField( 4 );
	gbl.setConstraints( batchPct, c );
	jc.add( batchPct );

	// add button
	c.gridwidth = GridBagConstraints.REMAINDER;
	jb = new JButton("Save Batch" );
	jb.setToolTipText( "Save training data in subsets of specified batch size." );
	jb.addActionListener( new ActionListener() {
	  public void actionPerformed(ActionEvent ae) {
		// Get the percentage from the textfields [make sure
		//   to strip '%' if existent].
		String batchPctStr = batchPct.getText();
		int batchSymIdx = batchPctStr.lastIndexOf("%");
		if (batchSymIdx != -1) {
		    batchPctStr = batchPctStr.substring(0,batchSymIdx-1);
		}
		double batchPercent = 0;
		try {
		    batchPercent =
			(new Double(batchPctStr)).doubleValue();
		    System.out.println("Batch percentage: " + batchPercent);
		} catch (NumberFormatException nfe) {
		    JOptionPane.showMessageDialog(DatasetEditor.this,
						  "Illegal percentage value.",
						  nfe.getMessage(),
						  JOptionPane.ERROR_MESSAGE);
		    return; // error.
		}
		// Convert percentages to doubles
		batchPercent /= 100;

		// now output to different files
		DatasetEditor.this.dset.batchOutput(getScenarios(training),
			      getScenarios(testing), batchPercent);
	  }
	});
	gbl.setConstraints( jb, c );
	jc.add( jb );

	// Done with the window contents.  Pack'n'show the window
	pack();
	setVisible( true );
	
    }

    /**
     * Gets Vectors of the  names of the Scenarios specified in the
     * Dataset.
     **/
    protected void getScenarioNameVectors()
    {
	training = getScenarioNameVector( dset.trainEnum() );
	testing = getScenarioNameVector( dset.testEnum() );
    }

    /**
     * Puts the names of the Scenarios in the Enumeration into the
     * given Vector. 
     *
     * @param     e An Enumeration<Scenario> to get the names from.
     * @return    A Vector of the Scenario names.
     **/
    protected Vector getScenarioNameVector( Enumeration e )
    {
	Scenario s;
	Vector v = new Vector();
	while ( e.hasMoreElements() ) {
	    s = (Scenario) e.nextElement();
	    v.addElement( s.getName() );
	} // end while over enumeration
	return v;
    }

    /**
     * Gets a Vector of Scenarios based on the given Vector of
     * Scenario Names.
     *
     * @param     names A Vector<String> of ScenarioNames.
     * @return    A Vector<Scenario> of the associated Scenarios.
     **/
    protected Vector getScenarios( Vector names )
    {
	Scenario s;
	Vector v = new Vector( names.size() );
	Enumeration e = names.elements();
	while ( e.hasMoreElements() ) {
	    v.addElement( sset.getScenario( (String)e.nextElement() ) );
	} // end while over Scenario names
	return v;
    }

    /**
     * Updates the percentages based on the data in the Vectors.
     **/
    public void updatePercentages()
    {
	trainingPct.setText(Fmt.dbl(((double)((double)training.size() /
					      (double)scenCt) * 100.0)) + "%");
	testingPct.setText(Fmt.dbl(((double)((double)testing.size() /
					     (double)scenCt) * 100.0)) + "%");
    }


    /**
     * Updates the Dataset with the values in the editing fields.
     **/
    public void updateDataset()
    {
	// Update the training Scenarios.
	dset.setTrainingScenarios( getScenarios( training ) );
	// Update the testing Scenarios.
	dset.setTestingScenarios( getScenarios( testing ) );
    }


    public void saveDatasetToFile()
      {
	//	String datasetString = dset.toString();
	// Allow the user to select a file name with which to save.
	String dir = "./data/datasets";
// 	  vlProps.getProperty("ntlsys.ntlhome") + "/" +
// 	    vlProps.getProperty("ntlsys.verblearn.homeDir") +
// 	      vlProps.getProperty("ntlsys.verblearn.dataDir") +
// 		vlProps.getProperty("ntlsys.verblearn.dsetDir");
	JFileChooser jfc = new JFileChooser( dir );
	//	int retval = jfc.showSaveDialog( VerbLearn.this );
	int retval = jfc.showSaveDialog( this );
	if (retval == JFileChooser.APPROVE_OPTION) {
	  final File toSave = jfc.getSelectedFile();
	  if (toSave != null) {
	    // add new Dataset to ScenarioSet (remain targeted to former one)
	    Dataset d = new Dataset(toSave.getName(), getScenarios(training),
				    getScenarios(testing));
	    sset.addDataset(d);

	    // Save the Dataset to the given
	    // filename, and do it on a seperate thread...
	    // and tell the user when done.
	    Thread t = new Thread() {
	      public void run() {
		saveDataset( toSave );
		    JOptionPane.showMessageDialog(DatasetEditor.this,
						  "Dataset has been saved.",
						  "Save Complete",
						  JOptionPane.INFORMATION_MESSAGE );
	      } // end Thread.run
	    }; // end Thead
	    t.start();
	  }
	}
      }


  // get rid of this -- it's in Datset.java as save; change there to saveDataset?
    public void saveDataset( File file) 
      {
	// Simple check
	if (file == null) {
	  JOptionPane.showMessageDialog(DatasetEditor.this,
					"No filename specified.",
					"No filename",
					JOptionPane.ERROR_MESSAGE);
	}

	if ( file.exists() ) {
	    if (JOptionPane.showConfirmDialog(DatasetEditor.this,
					      file.getAbsolutePath() +
					      " exists.  Replace?",
					      "File replacement confirmation",
					      JOptionPane.YES_NO_OPTION,
					      JOptionPane.WARNING_MESSAGE)
		== JOptionPane.NO_OPTION) {
		// They don't want to replace the file, so we stop here.
		return;
	    }
	}
	
	try {
	  PrintWriter pw = new PrintWriter(new FileWriter(file));
	  pw.write(dset.toString());
	  pw.close();
	}
	catch (IOException e) {
	  JOptionPane.showMessageDialog(DatasetEditor.this,
					"Failed to save dataset.",
					"Failed to save dataset.",
					JOptionPane.ERROR_MESSAGE);

	}
      }

    /**
     * An inner class to perform the action associated with the
     * 'add selected' button.
     **/
    class AddButtonActionListener
	implements ActionListener
    {
	private Vector v;
	private JList jl;
	public AddButtonActionListener(Vector v, JList jl)
	{
	    this.v = v;
	    this.jl = jl;
	}

	public void actionPerformed(ActionEvent ae)
	{
	    // Get the selected scenario names from the
	    // ScenarioList.
	    Object scNames[] = scenarioList.getSelectedValues();
	    for (int i=0; i<scNames.length; i++) {
		// Add them to the correct vector if they're not in
		// there already
		if (! v.contains( scNames[i] )) {
		    v.addElement( scNames[i] );
		}
	    } // end for over selected items
	    // Update the percentages.
	    updatePercentages();
	    // Update that vector's JList.
	    jl.setListData( v );
	    jl.validate();
	    // Redraw the visible protion of the list, since they
	    // don't seem to want to do it themselves.
	    jl.update( jl.getGraphics() );
	}
    } // end inner class AddButtonActionListener


    /**
     * An inner class to perform the action associated with the
     * 'remove selected' button.
     **/
    class RemoveButtonActionListener
	implements ActionListener
    {
	private Vector v;
	private JList jl;

	public RemoveButtonActionListener(Vector v, JList jl)
	{
	    this.v = v;
	    this.jl = jl;
	}


	public void actionPerformed(ActionEvent ae)
	{
	    Object toRemove[] = jl.getSelectedValues();
	    for (int i=0; i<toRemove.length; i++) {
		v.removeElement( toRemove[i] );
	    } // end for over selected items
	    // Update the percentages.
	    updatePercentages();
	    jl.setListData( v );
	    jl.validate();
	    // Redraw the visible protion of the list, since they
	    // don't seem to want to do it themselves.
	    jl.update( jl.getGraphics() );
	}
    } // end inner class RemoveButtonActionListener


}
