package NTLSystem.verblearn;

import com.sun.java.swing.JTextArea;
import java.io.*;
import java.util.*;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * A collection of Words [and their associated Senses], with
 * functionality for labelling, obeying, and learning.
 * <P>
 * <EM>This class needn't be public... it is only to get Javadoc 1.1
 * to generate docs for it.</EM>
 * <P>
 * $Id: Lexicon.java,v 1.7 1998/10/16 01:59:34 jsled Exp $<BR>
 * Last Editor: $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Word
 * @version   $Revision: 1.7 $
 **/
public class Lexicon
    implements Serializable
{

    // {{{ Attributes

    /** The serialization stream unique identifier. Recompute this
	[with serialver] if this class changes. **/
    static final long serialVersionUID = -4433917872821588938L;

    /**
     * This Lexicon's name.
     **/
    private String name;

    /**
     * The Language which this Lexicon is in.
     **/
    private String language;
    
    /**
     * The Words in this Lexicon.
     * StringSortedHashtable< String, Word >.  The keys are a String
     * represnetation of the Word's label [ VerbComplex.toString() ].
     **/
    private StringSortedHashtable words;

    /**
     * The ScenarioSet which contains this Lexicon.
     **/
    private ScenarioSet container;

    /**
     * A copy of the virtuals table, used for adaptation.  Maps String
     * Feature names to Doubles.
     **/
    private StringSortedHashtable virtuals;

    // }}} Attributes

    // {{{ Constructors 

    /**
     * Basic constructors for Lexicons.  Needed for Serialization.
     **/
    public Lexicon() {
	name = "New/Unnamed";
	language = "Unknown";
	words = new StringSortedHashtable();
	container = null;
	virtuals = new StringSortedHashtable();
    }

    /**
     * A better constructor for Lexicons.
     *
     * @param     name The name of this Lexicon.
     * @param     lang The Language of the entries in this Lexicon.
     * @param     sset The containing ScenarioSet.
     * @param     p The Params object to get virtual sample values
     *            from.
     **/
    public Lexicon(String name,
		   String lang,
		   ScenarioSet sset,
		   Params p)
    {
	this();
	this.name = name;
	language = lang;
	container = sset;
	// copy virtual sample values from p
	Enumeration e = p.featureNames();
	String fName;
	double vSamp;
	while ( e.hasMoreElements() ) {
	    try {
		fName = (String)e.nextElement();
		vSamp = p.getVirtualSamples( fName );
		setVirtualSamples( fName, vSamp );
	    } catch (NoSuchFeatureException nsfe) {
		// Shouldn't happen, since we're asking for Feature
		// names we were just told the Params object has.
		throw new InternalError( nsfe.getMessage() );
	    } // end try-catch
	} // end while over enumeration of vsamp Feature names
    }
    
    // }}} Constructors

    // {{{ Methods

    // {{{ name accessors

    /**
     * Get the name of this Lexicon.
     *
     * @return This Lexicon's name.
     **/
    public String getName() {return name;}
    
    /**
     * Set the name of this Lexicon.
     *
     * @param     newName The new name for this Lexicon.
     **/
    public void setName(String newName) {this.name = newName;}

    // }}} name accessors

    // {{{ language accessor

    /**
     * Gets the Language which this Lexicon is in.
     *
     * @return    The Language this Lexicon is in.
     **/
    public String getLanguage() { return language; }

    // }}} language accessor

    /**
     * Returns the number of Words in this Lexicon.
     *
     * @return    The number of Words in the Lexicon.
     **/
    public int wordCount() { return words.size(); }

    /**
     * Returns the number of Senses in this Lexcion.  Convenience
     * method.
     *
     * @return    The number of Senses over all Words in the Lexcion.
     **/
    public int senseCount()
    {
	int senseCount = 0;
	Enumeration e = getWords();
	Word w;
	while ( e.hasMoreElements() ) {
	    w = (Word)e.nextElement();
	    senseCount += w.senseCount();
	} // end while over e
	return senseCount;
    }

    /**
     * The number of examples which this Lexicon covers.
     *
     * @return    The total number of examples which this Lexicon has
     *            seen.
     **/
    public int numExamples()
    {
	int totalEx = 0;
	Enumeration e = getWords();
	while ( e.hasMoreElements() ) {
	    totalEx += ((Word)e.nextElement()).numExamples();
	} // end while over enumeration of Words
	return totalEx;
    }

    /**
     * Adds a Word to the Lexicon.
     *
     * @param     newWord The new Word to add to the Lexicon.
     **/
    protected void addWord(Word newWord)
    {
	words.put( newWord.getName(), newWord );
    }

    /**
     * Returns a particular Word in the Lexicon.
     *
     * @param     name The Word's 'name' [usually the string
     *            representation of the Word's label].
     * @return    The Word corresponding to that name, or null if it
     *            cannot be found.
     **/
    public Word getWord(String wordName)
    {
	return (Word)words.get( wordName );
    }
    
    /**
     * Returns an Enumeration of the Words in the Lexicon.  These are
     * returned in sorted order.
     *
     * @return     An Enumeration of the Words in the Lexicon in
     *             sorted order.
     **/
    public Enumeration getWords()
    {
	return new Enumeration() {
	    Enumeration sortedWords = words.getSortedKeys();
	    public boolean hasMoreElements() {
		return sortedWords.hasMoreElements();
	    }
	    public Object nextElement() {
		return words.get(sortedWords.nextElement());
	    }
	}; // end anonymous Enumeration
    }

    /**
     * Query to see if a given Word is in the Lexicon.  Uses
     * Word.equals() for checking.
     *
     * @param     w The Word to check for.
     * @return    True if the Word is in the Lexicon; False
     *            otherwise.
     **/
    public boolean contains(Word w)
    {
	return words.contains( w );
    }

    /**
     * Query to see if a given named Word is in the Lexicon.
     *
     * @param     wordName The name of the Word to check for.
     * @return    True if a Word with the given name is in the
     *            Lexicon; false otherwise.
     **/
    public boolean contains(String wordName)
    {
	return words.containsKey( wordName );
    }

    /**
     * Returns the ScenarioSet which contains this Lexicon.
     *
     * @return    The ScenarioSet which contains this Lexicon.
     **/
    public ScenarioSet getScenarioSet() { return container; }

    /**
     * Sets the ScenarioSet which contains this Lexicon.
     *
     * @param     sset The ScenarioSet which contains this Lexicon.
     **/
    public void setScenarioSet(ScenarioSet sset)
    { container = sset; }

    /**
     * Modify self to reflect a new labelled action.  This performs
     * model merging if necessary.
     *
     * @param     f  An Fstruct describing the new action.
     * @param     vc The label associated with that action [and thus
     *            that Fstruct].
     * @param     log The JTextArea to use for output.  If null, it
     *            won't be used.
     **/
    public void incorporate(Fstruct f,
			    VerbComplex vc,
			    JTextArea log)
    {

	// Optional (TBD):  Remove features explained by well-learned
	// words in the verb complex.
	
	// Incorporate remaining features paired with the verb complex
	// (TBD: include words which caused explaining-away above?)

	boolean ranMerge = false; 
	if ( contains( vc.toString() ) ) {
	    ranMerge = getWord( vc.toString() ).incorporate( f, log );
	} else {
	    Word newWord = new Word( vc, this );
	    ranMerge = newWord.incorporate( f, log );
	    addWord( newWord );
	}

	// Update the virtual samples
	if (getScenarioSet().getParams().adaptVirtuals() && ranMerge) {
	    adaptVirtuals();

	}
    }

    /**
     * Performs a merge loop if there have been any actions
     * incorporated since the last merge loop.  Used to "clean up" at
     * end of training. 
     *
     * @param     log The JTextArea for informational output.  If
     *            null, it won't be used.
     **/
    public void synch( JTextArea log ) {
	Enumeration wordsEnum = words.elements();
	while ( wordsEnum.hasMoreElements() ) {
	    ((Word)wordsEnum.nextElement()).synch( log );
	} // end while over wordsEnum

	if ( getScenarioSet().getParams().adaptVirtuals() ) {
	    adaptVirtuals();
	}
    }

    /**
     * Given a linking Fstruct, return the best-matching (multi-slot)
     * label. 
     *
     * @param     f An Fstruct describing an action.
     * @return    The VerbComplex label which best describes that
     *            action.
     **/
    public VerbComplex label(Fstruct f) {
	CandidateSense leading = new CandidateSense( (VerbComplex)null,
						     (Sense)null,
						     -1.0 );
	CandidateSense cur;
	Enumeration e = getWords();
	Word w;
	while (e.hasMoreElements()) {
	    w = (Word)e.nextElement();
	    cur = w.bestSenseFor( f, numExamples() );
	    if (cur.posterior > leading.posterior) {
		leading = cur;
		leading.label = w.getLabel();
	    }
	} // end while over enumeration
	return leading.label;
    }

    /**
     * For each feature, do a weighted average of (1) the current
     * settings, and (2) the numEx-averaged value of v = (1-p)/(pn-1)
     * across all senses (where n is the # of values, and p is the
     * probability of the mode value).
     **/
    private void adaptVirtuals() {
	// Iteratively adapt each feature:
	Enumeration fEnum = getVirtualSampleFeatures();
	Vector v = new Vector( virtuals.size() );
	while (fEnum.hasMoreElements()) {
	    v.addElement( (String)fEnum.nextElement() );
	} // end while over virtual sample feature name enumeration
	for (int i=0; i<v.size(); i++) {
	    String fname = (String)v.elementAt( i );
	    double vAvg = 0;
	    double count = 0;
	    // Double loop to average mode's probability over all
	    // senses in slot:
	    Enumeration wEnum = getWords();
	    Word w;
	    while (wEnum.hasMoreElements()) {
		w = (Word)wEnum.nextElement();
		Enumeration sEnum = w.getSenses();
		while (sEnum.hasMoreElements()) {
		    Sense s = (Sense)sEnum.nextElement();
		    // NOTE: We're assured that this is a
		    // DiscreteDist, because that's all that VerbLearn
		    // deals with.  Actually, it's a
		    // VirtualSampleDiscreteDist, but DiscretDist is
		    // shorter, and polymorphism will provide the
		    // methods we need. :) 
		    try {
			DiscreteDist dd =
			    (DiscreteDist)s.getFeature(fname).getValue();
			if (dd != null) { 
			    double p = dd.getProb( dd.mode().doubleValue() );
			    // Avoid infinity when adaptVirtuals
			    // encounters a uniform dist: 
			    vAvg += s.numExamples() * 
				Math.min(getScenarioSet().getParams().maxVirtual(),
					 (1 - p) / (dd.getSize() * p - 1));
			count += s.numExamples();
			}
		    } catch (NTLException ntle) {
			    // Shouldn't happen as all Senses should have
			// the same Features.
			throw new InternalError( ntle.getMessage() );
		    }
		}
	    }
	    // Weight the above against the current virtuals setting:
	    double virtInertia = getScenarioSet().getParams().virtualInertia();
	    vAvg = vAvg + virtInertia *
		((Double)virtuals.get(fname)).doubleValue();
	    // assertion
	    if ( count != numExamples() ) {
		System.out.println("Lexicon.adaptVirtuals(): count ("
				   + count + ") != numExamples (" +
				   numExamples() + ").");
	    }
	    count = count + virtInertia;
	    vAvg = (double)((double)vAvg / (double)count);
	    // Finally set the new value:
	    virtuals.put(fname, new Double(vAvg));
	}
    }

    /**
     * Returns the Feature names for which a virtual sample count
     * exists. 
     *
     * @return    The virtual sample keys. FWIW, these are returned in
     *            name-sorted order.
     **/
    public Enumeration getVirtualSampleFeatures()
    {
	return virtuals.getSortedKeys();
    }

    /**
     * Returns the virtual sample count associated with the given
     * Feature name.
     *
     * @param     featureName The name of the Feature for which
     *            to return the virtual sample count.
     * @return    The virtual sample count for the given Feature.
     **/
    public double getVirtualSamples( String featureName )
    {
	Double toRet = (Double)virtuals.get( featureName );
	if ( toRet == null ) {
	    return 0.0;
	}
	return toRet.doubleValue();
    }

    /**
     * Sets the virtual sample count for the given Feature name.
     *
     * @param     featureName The name of the Feature for which to set
     *            the virtual sample count.
     * @param     virtSamp The virtual sample count for the value.
     **/
    public void setVirtualSamples( String featureName,
				   double virtSamp )
    {
	virtuals.put( featureName,
		      new Double( virtSamp ) );
    }


    /**
     * A more detailed version of label(), this routine returns a
     * Vector of the posterior probabilities on all senses in the
     * Lexicon.
     *
     * @param     f The linking Fstruct to label.
     * @return    A Vector<CandidateSense> containing the posterior
     *            probabilities of all Senses in the Lexicon for the
     *            given Fstruct.
     **/
    public Vector allSensePosteriors(Fstruct f) {
	Vector toRet = new Vector( senseCount() );
	Enumeration e = getWords();
	Enumeration cEnum;
	Word curWord;
	while ( e.hasMoreElements() ) {
	    curWord = (Word)e.nextElement();
	    //	    cEnum =
	    //curWord.allSensePosteriors(f, curWord.senseCount()).
	    //elements();
	    cEnum = curWord.allSensePosteriors(f, numExamples()).elements();
	    while ( cEnum.hasMoreElements() ) {
		toRet.addElement( cEnum.nextElement() );
	    } // end while over CandidateSense enumeration
	} // end while over Word enumeration
	return toRet;
    }


    /**
     * Return a linking Fstruct compatible with a command and initial
     * state linking features.
     *
     * @param     cmd The label/name of the action.
     * @param     state The initial linking Fstruct.
     * @return    An Fstruct describing the state after the action has
     *            been obeyed.
     **/
    public Fstruct obey(VerbComplex cmd, Fstruct state)
    {
	Fstruct toRet;
	if (cmd == null) {
	    return null;
	}
	Word w = (Word)words.get( cmd.toString() );
	if (w == null) {
	    toRet = (Fstruct)state.clone();
	} else {
	    CandidateSense cand = w.bestSenseFor( state, w.numExamples() );
	    if (cand.posterior >= getScenarioSet().getParams().minObey()) {
		// Get a clone of the Sense Fstruct so we can remove
		// Features at will.
		toRet = (Fstruct)cand.sense.clone();
		// Override the Features in state...
		try {
		    Vector stateFeatures =
			state.getFeatures( state.getAllFeatureNames().
					   elements() );
		    // Go through the Vector, replacing each element
		    // with it's clone.
		    Feature f;
		    for (int i=0; i<stateFeatures.size(); i++) {
			f = (Feature)stateFeatures.elementAt(i);
			toRet.removeFeature( f.getName() );
			stateFeatures.setElementAt(f.clone(), i);
		    } // end for over state Features
		    // Add these Features to the to-return Fstruct.
		    toRet.putFeatures( stateFeatures.elements() );
		} catch (NTLException ntle) {
		    // We expect neither, because we're giving
		    // getFeatures(...) Strings, we're gettings
		    // Features it has, and we're giving putFeatures
		    // Features.
		    throw new InternalError( ntle.getMessage() );
		} // end try-catch
	    } else {
		toRet = (Fstruct)state.clone(); 
	    } 
	}

	// Remove the non-peaked-enough values before continuing.
	// This, before [ie: in dbailey's VerbLearn] was handled by an
	// Fstruct constructor over an array of 'bestSenseFor'
	// Fstructs from ea. Slot.  Since Slots are gone, this needs
	// to be done explicitly.
	toRet.removeNonPeakedFeatures( getScenarioSet().getParams().
				       minSetFeature() );
	return toRet;
    }

    public String toString() {
	StringBuffer res = new StringBuffer();
	res.append("Lexicon " + getName() + " (" +
		   senseCount() + " senses):\n");
	Enumeration wordsEnum = words.elements();
	while ( wordsEnum.hasMoreElements() ) {
	    res.append( ((Word)wordsEnum.nextElement()).toString() );
	} // end while over words
	return res.toString();
    }

    // }}} Methods
}
