package NTLSystem.verblearn

import java.io.*;
import java.util.*;
import NTLSystem.support.*;

///////////////////////////////////////////////////////////////////////
// Discrete probability distribution

class Dist implements Serializable {
  
  private String id;  // holds the feature name

  private Hashtable counts;  // maps String feature names to Double
                             // These Double wrappers are horribly inefficient.

  private double total;  // sum of counts (WITHOUT virtual samples added)

  private double virt; // The number of virtual samples which should be
                       // implicitly added to each count.

  private double kVirt;  // Cached value of virt * number of values
                         // for this feature

  // Construct a new distribution with 'virt' virtual samples for each value
  // of feature 'fname'.
  Dist(String fname, double virt) {
    id = fname;
    this.virt = virt;
    kVirt = VerbLearn.context().params().featureVals(fname).size() * virt;
    counts = new Hashtable();
    Enumeration valEnum = VerbLearn.context().params()
      .featureVals(fname).elements();
    while (valEnum.hasMoreElements()) {
      setCount((String)valEnum.nextElement(), 0);
    }
  }

  // Construct a new distribution by merging two existing
  // distributions.  NOTE: This version does not add virtual sample
  // component of each distribution, unlike Sather version.  That
  // version did not allow distributions to grow more peaked with
  // further training examples.  Instead, we use the current virtual
  // sample count of the slot, passed in 'virt'.  This algorithm still
  // needs its mathematical justification worked out.
  Dist(String fname, Dist d1, Dist d2, double virt) {
    id = fname;
    // if (d1.virt != d2.virt) {
    //   System.err.println("WARNING: Merging distributions with " +
    //  		    "unequal virtuals\n");
    // }
    this.virt = virt;
    kVirt = VerbLearn.context().params().featureVals(fname).size() * virt;
    counts = new Hashtable();
    Enumeration valEnum = d1.counts.keys();
    while (valEnum.hasMoreElements()) {
      String valName = (String)valEnum.nextElement();
      setCount(valName, d1.getCount(valName) + d2.getCount(valName));
    }
    total = total + d1.total + d2.total;
  }

  // Used to account for a new example of the feature set to "vname".
  public void incrementValue(String vname) {
    setCount(vname, getCount(vname) + 1);
    total = total + 1;
  }

  // Return the probability of a particular value for the feature..
  public double likelihood(String val) { 
    return (getCount(val) + virt) / (total + kVirt);
  }

  // Return the highest-probability feature value for this distribution.
  public String mode() { 
    String leadingVal = null;
    double leadingCount = -1;
    Enumeration valEnum = counts.keys();
    while (valEnum.hasMoreElements()) {
      String val = (String)valEnum.nextElement();
      double count = getCount(val);
      if (count > leadingCount) {
	leadingCount = count;
	leadingVal = val;
      }
    }
    return leadingVal;
  }

  // Return the second-highest-probability feature value for this 
  // distribution (or null if there is only one feature in the dist).
  private String runnerUp() { 
    String leadingVal = null;
    double leadingCount = -1;
    String runnerUpVal = null;
    double runnerUpCount = -1;
    Enumeration valEnum = counts.keys();
    while (valEnum.hasMoreElements()) {
      String val = (String)valEnum.nextElement();
      double count = getCount(val);
      if (count > leadingCount) {
	runnerUpVal = leadingVal;
	runnerUpCount = leadingCount;
	leadingVal = val;
	leadingCount = count;
      }
      else if (count > runnerUpCount) {
	runnerUpVal = val;
	runnerUpCount = count;
      }
    }
    return runnerUpVal;
  }

  // A heuristic measure of the peakedness of the distribution.
  // Currently we use the ratio of the likelihood of the mode to
  // the likelihood of the runner-up to the mode.
  // Thus, peakedness ranges from 1.0 for a uniform distribution,
  // up to infinity for a spike distribution.
  public double peakedness() { 
    String ru = runnerUp();
    if (ru == null) return Double.POSITIVE_INFINITY;
    return likelihood(mode()) / likelihood(ru);
  }

  // A heuristic measure of the difference between two distributions.
  // Basically chi-square, except we use a normalized version to
  // factor out differences in number of samples, and then invert it
  // with 1/(1+chi) to give a measure ranging from 0 (very different)
  // to 1 (identical).
  public double similarityTo(Dist d) {
    double res = 1 / (1 + normalizedChiSquareWith(d));
    return res;
  }

  // From Numerical Recipes 2nd ed., p. 623 ("r" and "s" in that
  // alg. map to "this" and "d" respectively).
  // Probabilities rather than counts are used, since otherwise
  // chi square is problematic because it also measures differences 
  // in the number of samples in the two distributions.  
  public double normalizedChiSquareWith(Dist d) {
    double res = 0;
    Enumeration iEnum = counts.keys();
    while (iEnum.hasMoreElements()) {
      String i = (String)iEnum.nextElement();
      double ri = likelihood(i);
      double si = d.likelihood(i);
      if (ri != 0 || si != 0) {
	res = res + Math.pow(ri - si, 2) / (ri + si);
      }
    }
    return res;
  }

  // This is used locally to reduce casting clutter.
  private double getCount(String val) {
    return ((Double)counts.get(val)).doubleValue();
  }

  // This is used locally to reduce casting clutter.
  private void setCount(String val, double count) {
    counts.put(val, new Double(count));
  }

  public String toString() {
    StringBuffer res = new StringBuffer();
    String m = mode();
    Enumeration valEnum = VerbLearn.context().params()
      .featureVals(id).elements();
    while (valEnum.hasMoreElements()) {
      String valName = (String)valEnum.nextElement();
      if (res.length() != 0) res.append(" ");
      // For ease of reading, capitalize the mode value (if peaked enough):
      if (valName.equals(m) && 
	  peakedness() >= VerbLearn.context().params().minSetFeature()) {
	res.append(valName.toUpperCase());
      }
      else {
	res.append(valName);
      }
      res.append("=").append(Fmt.dbl(likelihood(valName)));
      // FOR COUNTS INSTEAD OF PROBABILITIES, USE:
      // ((Double)counts.get(valName)).doubleValue() + virt
    }
    return res.toString();
  }
}
