package NTLSystem.verblearn;

import java.io.*;
import java.util.*;
import NTLSystem;
import NTLSystem.support.*;

/**
 * A collection of Words [and their associated Senses], with
 * functionality for labelling, obeying, and learning.
 * <P>
 * <EM>This class needn't be public... it is only to get Javadoc 1.1
 * to generate docs for it.</EM>
 * <P>
 * $Id: Lexicon.java,v 1.2 1998/07/31 21:11:27 jsled Exp $<BR>
 * Last Editor: $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Word
 * @version   $Revision: 1.2 $
 **/
public class Lexicon
    implements Serializable
{

    // {{{ Attributes

    /**
     * This Lexicon's name.
     **/
    private String name;

    /**
     * The Language which this Lexicon is in.
     **/
    private String language;
    
    /**
     * The Words in this Lexicon.  Vector<Word>.
     **/
    private StringSortedHashtable words;

    /**
     * The ScenarioSet which contains this Lexicon.
     **/
    private ScenarioSet container;

    /**
     * A copy of the virtuals table, used for adaptation.  Maps String
     * Feature names to Doubles.
     **/
    public Hashtable virtuals;

    // }}} Attributes

    // {{{ Constructors 

    /**
     * Basic constructors for Lexicons.  Needed for Serialization.
     **/
    public Lexicon() {
	name = "New/Unnamed";
	language = "Unknown";
	words = new StringSortedHashtable();
	container = null;
    }

    /**
     * A better constructor for Lexicons.
     *
     * @param     name The name of this Lexicon.
     * @param     lang The Language of the entries in this Lexicon.
     * @param     sset The containing ScenarioSet.
     **/
    public Lexicon(String name, String lang, ScenarioSet cont)
    {
	this();
	this.name = name;
	language = lang;
	container = cont;
    }
    
    // }}} Constructors

    // {{{ Methods

    // {{{ name accessors

    /**
     * Get the name of this Lexicon.
     *
     * @return This Lexicon's name.
     **/
    public String getName() {return name;}
    
    /**
     * Set the name of this Lexicon.
     *
     * @param     newName The new name for this Lexicon.
     **/
    public void setName(String newName) {this.name = newName;}

    // }}} name accessors

    // {{{ language accessor

    /**
     * Gets the Language which this Lexicon is in.
     *
     * @return    The Language this Lexicon is in.
     **/
    public String getLanguage() { return language; }

    // }}} language accessor

    /**
     * Returns the number of Words in this Lexicon.
     *
     * @return    The number of Words in the Lexicon.
     **/
    public int wordCount() { return words.size(); }

    /**
     * Returns the number of Senses in this Lexcion.  Convenience
     * method.
     *
     * @return    The number of Senses over all Words in the Lexcion.
     **/
    public int senseCount()
    {
	int senseCount = 0;
	Enumeration e = getWords();
	Word w;
	while ( e.hasMoreElements() ) {
	    w = (Word)e.nextElement();
	    senseCount += w.senseCount();
	} // end while over e
	return senseCount;
    }

    /**
     * The number of examples which this Lexicon covers.
     *
     * @return    The total number of examples which this Lexicon has
     *            seen.
     **/
    public int numExamples()
    {
	int totalEx = 0;
	Enumeration e = getWords();
	while ( e.hasMoreElements() ) {
	    totalEx += ((Word)e.nextElement()).numExamples();
	} // end while over enumeration of Words
	return totalEx;
    }

    /**
     * Adds a Word to the Lexicon.
     *
     * @param     newWord The new Word to add to the Lexicon.
     **/
    protected void addWord(Word newWord)
    {
	words.put( newWord.getName(), newWord );
    }

    /**
     * Returns a particular Word in the Lexicon.
     *
     * @param     name The Word's 'name' [usually the same as the
     *            Word's label].
     * @return    The Word corresponding to that name, or null if it
     *            cannot be found.
     **/
    public Word getWord(String wordName)
    {
	return words.get( wordName );
    }
    
    /**
     * Returns an Enumeration of the Words in the Lexicon.
     *
     * @return     An Enumeration of the Words in the Lexicon.
     **/
    public Enumeration getWords() { return words.elements(); }

    /**
     * Query to see if a given Word is in the Lexicon.  Uses
     * Word.equals() for checking.
     *
     * @param     w The Word to check for.
     * @return    True if the Word is in the Lexicon; False
     *            otherwise.
     **/
    public boolean contains(Word w)
    {
	return words.contains( w );
    }

    /**
     * Query to see if a given named Word is in the Lexicon.
     *
     * @param     wordName The name of the Word to check for.
     * @return    True if a Word with the given name is in the
     *            Lexicon; false otherwise.
     **/
    public boolean contains(String wordName)
    {
	return words.containsKey( wordName );
    }

    /**
     * Returns the ScenarioSet which contains this Lexicon.
     *
     * @return    The ScenarioSet which contains this Lexicon.
     **/
    public ScenarioSet getContainer()
    {
	return container;
    }

    /**
     * Modify self to reflect a new labelled action.  This performs
     * model merging if necessary.
     *
     * @param     f  An Fstruct describing the new action.
     * @param     vc The label associated with that action [and thus
     *            that Fstruct].
     **/
    public void incorporate(Fstruct f, VerbComplex vc) {

	// Optional (TBD):  Remove features explained by well-learned
	// words in the verb complex.
	
	// Incorporate remaining features paired with the verb complex
	// (TBD: include words which caused explaining-away above?)

	// FIXME: ignore this loop over slots, and incorporate into
	// the correct Word [creating if non-existant]?  Check.
	/*
	for (int i = 0; i < slots.length; i++) {
	    if (vc.slotFilled(i)) {
		slots[i].incorporate(f, vc.getWord(i));
	    }
	}
	*/
	// ^^^^ -- OLD -- ^^^^
	// vvvv -- NEW -- vvvv
	
	if ( isContained( vc.toString() ) ) {
	    getWord( vc.toString() ).incorporate( f );
	} else {
	    Word newWord = new Word( vc.toString(), this );
	    newWord.incorporate( f );
	    addWord( newWord );
	}
    }

    /**
     * Performs a merge loop if there have been any actions
     * incorporated since the last merge loop.  Used to "clean up" at
     * end of training. 
     **/
    public void synch() {
	for (int i = 0; i<words.size(); i++) {
	    ((Word)words.elementAt(i)).synch();
	}
    }

    /**
     * Given a linking Fstruct, return the best-matching (multi-slot)
     * label. 
     *
     * @param     f An Fstruct describing an action.
     * @return    The VerbComplex label which best describes that
     *            action.
     **/
    public VerbComplex label(Fstruct f) {
	CandidateSense leading = new CandidateSense("none", null, -1);
	Cnadidate Sense cur;
	Enumeration e = getWords();
	Word w;
	while (e.hasMoreElements()) {
	    w = (Word)e.nextElement();
	    cur = w.bestSenseFor(f, numEx);
	    if (cur.posterior > leading.posterior) {
		leading = cur;
	    }
	} // end while over enumeration
	return leading.label;

	// ^^^^ -- NEW -- ^^^^
	// vvvv -- OLD -- vvvv

	// Copy f so that features only get explained away in the local context!
	f = f.clone();
	VerbComplex vc = new VerbComplex(VerbLearn.context().numSlots());
	int numFilled = 0;
	CandidateSense[] candidates =
	    new CandidateSense[VerbLearn.context().numSlots()];
	
	// Repeatedly get the best fillers for each unfilled slot and choose
	// the best one if its posterior >= minLabel:
	while (true) {
	    
	    // Get the best label for each unfilled slot:
	    for (int i = 0; i < slots.length; i++) {
		candidates[i] = null;
		if (vc.getWord(i) == null) {
		    candidates[i] = slots[i].label(f);
		}
	    }

	    // Find the slot with MAP label:
	    int winningSlot = -1;
	    for (int i = 0; i < slots.length; i++) {
		if (vc.getWord(i) == null && 
		    candidates[i] != null &&
		    (winningSlot == -1 ||
		     candidates[i].posterior >= candidates[winningSlot].posterior)) {
		    winningSlot = i;
		}
	    }

	    // If the match is good enough, apply the label.
	    // Note this allows the possibility of producing an empty label
	    // meaning "I can't describe this".
	    if (winningSlot == -1 || 
		candidates[winningSlot].posterior 
		< VerbLearn.context().params().minLabel()) {
		break;
	    }
	    vc.setWord(winningSlot, candidates[winningSlot].label);
	    numFilled += 1;
	    if (numFilled == VerbLearn.context().numSlots()) { 
		break;
	    }
      
	    // Remove explained features from `f' before we go back and 
	    // embellish our label by filling further slots.
	    f.explainAway( candidates[winningSlot].sense );
	}
	
	// Finally, return the verb complex:
	return vc;
    }

    /**
     * For each feature, do a weighted average of (1) the current
     * settings, and (2) the numEx-averaged value of v = (1-p)/(pn-1)
     * across all senses (where n is the # of values, and p is the
     * probability of the mode value).
     **/
    private void adaptVirtuals() {
	// Iteratively adapt each feature:
	Enumeration fEnum = virtuals.keys();
	while (fEnum.hasMoreElements()) {
	    String fname = (String)fEnum.nextElement();
	    double vAvg = 0;
	    double count = 0;
	    // Double loop to average mode's probability over all senses in slot:
	    Enumeration wEnum = getWords();
	    while (wEnum.hasMoreElements()) {
		Enumeration sEnum = ((Word)wEnum.nextElement()).getSenses();
		while (sEnum.hasMoreElements()) {
		    Sense s = (Sense)sEnum.nextElement();
		    // FIXME:
		    // p = ((ProbDist)s.getFeature(fname).
		    //                  getValue()).getMLValue();
		    double p = s.getDist(fname).likelihood(s.getDist(fname).mode());
		    // Avoid infinity when adaptVirtuals encounters a uniform dist:
		    vAvg = vAvg + s.numEx() * 
			Math.min(VerbLearn.context().params().maxVirtual(),
				 (1 - p) / 
				 (p * VerbLearn.context().params().
				  featureVals(fname).size() - 1));
		    count = count + s.numEx();
		}
	    }
	    // Weight the above against the current virtuals setting:
	    vAvg = vAvg + VerbLearn.context().params().virtualInertia() *
		((Double)virtuals.get(fname)).doubleValue();
	    count = count + VerbLearn.context().params().virtualInertia();
	    vAvg = vAvg / count;
	    // Finally set the new value:
	    virtuals.put(fname, new Double(vAvg));
	}
    }
    

    /**
     * A more detailed version of label(), this routine returns a
     * Vector of the posterior probabilities on all senses in the
     * Lexicon.
     *
     * @param     f The linking Fstruct to label.
     * @return    A Vector<CandidateSense> containing the posterior
     *            probabilities of all Senses in the Lexicon for the
     *            given Fstruct.
     **/
    public Vector allSensePosteriors(Fstruct f) {
	Vector toRet = new Vector( senseCount() );
	Enumeration e = getWords();
	Enumeration cEnum;
	while ( e.hasMoreElements() ) {
	    cEnum = ((Word)e.nextElement()).allSensePosteriors(f);
	    while ( cEnum.hasMoreElements() ) {
		toRet.addElement( cEnum.nextElement() );
	    } // end while over CandidateSense enumeration
	} // end while over Word enumeration

	return toRet;

	// ^^^^ -- NEW -- ^^^^
	// vvvv -- OLD -- vvvv

	for (int i = 0; i < slots.length; i++) {
	    Vector cands = slots[i].allSensePosteriors(f);
	    Enumeration cEnum = cands.elements();
	    while (cEnum.hasMoreElements()) {
		res.append((CandidateSense)cEnum.nextElement());
	    }
	}
	return res.toString();
    }


    /**
     * Return a linking Fstruct compatible with a command and initial
     * state linking features.
     *
     * @param     cmd The label/name of the action.
     * @param     state The initial linking Fstruct.
     * @return    An Fstruct describing the state after the action has
     *            been obeyed.
     **/
    public Fstruct obey(VerbComplex cmd, Fstruct state)
    {
	if (cmd == null) return null;
	Word w = (Word)words.get(cmd);
	if (w == null) return null;
	CandidateSense cand = w.bestSenseFor(state, w.numExamples());
	if (cand.posterior >= VerbLearn.context().params().minObey()) {
	    return cand.sense;
	} else {
	    return null;
	}

	// ^^^^ -- NEW -- ^^^^
	// vvvv -- OLD -- vvvv

	// Conflict avoidance:  Choose minimally conflicting senses of the
	// various words in the command.  For now this is done by choosing
	// the sense for each word which best matches the initial world state.
	// There is no inter-word conflict avoidance.
	Sense[] winningSenses = new Sense[slots.length];
	for (int i = 0; i < slots.length; i++) {
	    winningSenses[i] = slots[i].obey(cmd.getWord(i), state);
	}
	
	// Conflict resolution:  Merge the chosen senses into a linking Fstruct.
	// For each motoric linking feature, choose MAP value from most
	// peaked distribution.  (This is actually implemented in Fstruct.)
	return new Fstruct(winningSenses, state);
    }

    public String toString() {
	StringBuffer res = new StringBuffer();
	res.append("Lexicon " + id + " (" + totalSenses() + " senses):\n");
	for (int i = 0; i < slots.length; i++) {
	    res.append(slots[i]);
	}
	return res.toString();
    }

    // }}} Methods
}
