package NTLSystem.verblearn;

import com.sun.java.swing.*;
import com.sun.java.swing.event.*;
import com.sun.java.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyVetoException;
import java.util.Enumeration;
import java.util.Vector;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * This is an InternalFrame which is used for viewing a single
 * Scenario within a ScenarioSet.  As well, from this InternalFrame,
 * all the Scenarios can be accessed, and selections from the
 * ScenarioSet can be made.  After a selection is made, the
 * ScenarioViewer is retargeted [through calling the 'setTarget()'
 * method] to the new Scenario, and it can be viewed.
 * <P>
 * <EM>This class needn't be public... it is only to get Javadoc 1.1
 * to generate docs for it.</EM>
 * <P>
 * $Id: ScenarioViewer.java,v 1.6 1998/10/16 01:59:42 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       ScenarioSet
 * @see       Scenario
 * @see       VerbLearn
 * @version   $Revision: 1.6 $
 **/

public class ScenarioViewer
    extends JInternalFrame
    implements ListSelectionListener
{
  private static String frameTitleStem = "Scenario Viewer";

    /**
     * The ScenarioSet which we're responsible for editing.
     **/
    protected ScenarioSet sset;
    /**
     * The Scenario from the ScenarioSet which we're currently
     * viewing.
     **/
    protected Scenario sc;
    /**
     * A collection of Boolean values indicating language-label
     * display toggle.  This should be the same size as the
     * ScenarioSet's language Vector.
     **/
    protected Vector langLabelDisplay;
    /**
     * The list of scenario names which this viewer can view.
     **/
    private JList scenarioList;
    /**
     * The name of this scenario.
     **/
    private JTextField nameTF;
    /**
     * The panel [which is inside a JScrollPane] which contains the
     * Fstruct viewers.
     **/
    private JPanel fstructs;
    /**
     * The table used for labeling.
     **/
    private JTable labelTable;
    /**
     * The ScenarioTableModel used by the labelTable.
     **/
    private ScenarioTableModel stm;
    /**
     * The language combo box
     **/
    private JComboBox langsCB;
    /**
     * The Selectometer used for displaying selections.
     **/
    private SelectoMeter sm;

    /*
     * Actions which can be performed in the ScenarioViewer
     */

    /**
     * Action to target the next Scenario in the ScenarioList. This
     * is basically convenience for clicking on the list itself.
     **/
    private Action nextScenario;
    /**
     * Action to target the previous Scenario in the
     * ScenarioList. This is basically convenience for clicking on the
     * list itself. 
     **/
    private Action prevScenario;
    private Action nextSelectedScenario;
    private Action prevSelectedScenario;
    private Action selectScenarios;
    private Action copyAndAdd;
    private Action deleteSelected;

    /**
     * Creates a ScenarioViewere object initialized to the given
     * ScenarioSet, and using the default value of the first Scenario
     * object within that ScenarioSet.
     *
     * @param     sset The ScenarioSet which is to be edited.
     **/
    public ScenarioViewer( ScenarioSet sset )
    {
	this( sset, null );
    }

    /**
     * Creates a ScenarioViewer object initialized to the given
     * ScenarioSet and the given Scenario object within that
     * ScenarioSet.
     *
     * @param     sset The ScenarioSet which is to be edited.
     * @param     scen The Scenario which is to be edited.  If null,
     *            the first Scenario from the ScenarioSet will be used
     *            to start with.
     **/
  public ScenarioViewer( ScenarioSet ss, Scenario scen )
  {
    super(frameTitleStem, true, true, true, true);

    sset = ss;
    sc = scen;

    // Set up the language label display vector for the labeling
    // table.
    Vector langs = sset.getLanguages();
    langLabelDisplay = new Vector( langs.size() );
    for (int i=0; i<langs.size(); i++) {
	langLabelDisplay.addElement( new Boolean( true ) );
    }

    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();

    JComponent cp = (JComponent)getContentPane();
    JButton jb;

    cp.setLayout(gbl);

    c.fill = GridBagConstraints.BOTH;
    c.gridwidth = 1;
    c.gridheight = 2;
    c.weightx = 0.15;
    scenarioList = new JList( sset.scenarios );
    scenarioList.setFixedCellWidth( 150 );
    // This list is single-selection only.
    scenarioList.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
    // Selections in the list should call 'setTarget( Scenario )'.
    scenarioList.addListSelectionListener( this );
    JScrollPane jsp = new JScrollPane( scenarioList );
    
    gbl.setConstraints(jsp, c);
    cp.add(jsp);

    c.gridheight = 1;
    JPanel jp = new JPanel();

    JLabel jl = new JLabel("Name:");
    jp.add(jl);

    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridwidth = GridBagConstraints.REMAINDER;
    nameTF = new JTextField( "<untitled>", 20 );
    jp.add(nameTF);
    gbl.setConstraints(jp, c);
    cp.add(jp);

    // Add the Fstructs for this scenario...
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.fill = GridBagConstraints.BOTH;
    c.weightx = 0.85;
    c.weighty = 1.0;
    fstructs = new JPanel( new GridBagLayout() );
    // Add the fstruct viewers to the panel.
    FstructViewer fsv;
    for (int i=0; i<sset.getTemplateFstructs().size(); i++) {
	fsv = new FstructViewer( null );
	fsv.setBorder( BorderFactory.createLineBorder( Color.black ) );
	((GridBagLayout)fstructs.getLayout()).
	    setConstraints( fsv, FstructViewer.panelConstraints );
	fstructs.add( fsv );
    } // end for over i

    jsp = new JScrollPane(fstructs,
			  ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
			  ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
    Dimension size = new Dimension( 700, 450 );
    jsp.setPreferredSize( size );
    jsp.setMaximumSize( size );
    gbl.setConstraints( jsp, c );
    cp.add(jsp);

    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.weightx = c.weighty = 0.0;

    // table model is, by inner-class-reference-nature, targeted to
    // current scenario...
    stm = new ScenarioTableModel();
    jp = new JPanel();
    jp.setBorder(BorderFactory.createTitledBorder("Scenario Label"));

    labelTable = new JTable( stm );
    // Compute the height for 3 visible rows
    int height = 3 * labelTable.getRowHeight();
    Dimension tableDim = new Dimension(400, height);
    labelTable.setPreferredScrollableViewportSize( tableDim );
    // Setup a combo box for the label column editor
    TableColumn labelColumn = labelTable.getColumnModel().getColumn( 2 );
    labelColumn.setCellEditor( new LabelTableCellEditor() );
    // Set the Table to not auto-create columns from the model, from
    // now on. This will prevent the table from resizing the columns
    // and forgetting about the column editor [the custom CellEditor
    // inner-class below] when fireTableStructureChanged() is called
    // in the ScenarioTableModel, below.
    labelTable.setAutoCreateColumnsFromModel( false );
    // Put the table in a ScrollPane
    jsp = new JScrollPane( labelTable );
    jsp.setMinimumSize( tableDim );
    jsp.setPreferredSize( tableDim );
    jp.add(jsp);

    gbl.setConstraints(jp, c);
    cp.add( jp );

    JPanel lowerPanel = new JPanel();

    jl = new JLabel("Default Language:");
    lowerPanel.add(jl);

    langsCB = new JComboBox( new VectorComboBoxModel(sset.getLanguages()) );
    lowerPanel.add( langsCB );

    // Create a sub-panel with the SelectoMeter and the Scene-Viewing buttons.
    GridBagLayout gbl2 = new GridBagLayout();
    jp = new JPanel(gbl2);
    c.gridheight = 1;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.fill = GridBagConstraints.HORIZONTAL;
    c.weightx = 0.0;
    c.anchor = GridBagConstraints.CENTER;
    // FIXME:
    sm = new SelectoMeter();
    gbl2.setConstraints(sm, c);
    jp.add(sm);

    c.gridwidth = 1;
    jb = new JButton("<<");
    jb.setToolTipText("Previous selected");
    // FIXME:
    // jb.addActionListener(prevSelectedScenario);
    // setTarget( ... );
    gbl2.setConstraints(jb, c);
    jp.add(jb);

    jb = new JButton("<");
    jb.setToolTipText("Previous");
    // FIXME:
    prevScenario = new SelectionAction() {
	public void actionPerformed(ActionEvent ae) {
	    super.actionPerformed( ae );
	    if ( sset.scenarios.getSize() > 0 ) {
		int idx = scenarioList.getSelectedIndex();
		if (idx == -1) { return; }
		idx = (idx-1 >= 0) ? idx-1 : idx;
		scenarioList.
		setSelectedValue( scenarioList.getModel().
				  getElementAt( idx ), true );
	    }
	}
    };
    jb.addActionListener(prevScenario);
    // setTarget( ... );
    gbl2.setConstraints(jb, c);
    jp.add(jb);

    jb = new JButton("Select");
    // FIXME:
    // jb.addActionListener(selectScenarios);
    // setTarget( ... );
    gbl2.setConstraints(jb, c);
    jp.add(jb);

    c.gridwidth = GridBagConstraints.RELATIVE;
    jb = new JButton(">");
    jb.setToolTipText("Next");
    nextScenario = new SelectionAction() {
	public void actionPerformed(ActionEvent ae) {
	    super.actionPerformed( ae );
	    int idx = scenarioList.getSelectedIndex();
	    if (idx == -1) { return; }
	    idx = (idx+1 <= sset.scenarios.getSize()-1) ? idx+1 : idx;
	    scenarioList.setSelectedIndex( idx );
	}
    };
    jb.addActionListener(nextScenario);
    gbl2.setConstraints(jb, c);
    jp.add(jb);

    c.gridwidth = GridBagConstraints.REMAINDER;
    jb = new JButton(">>");
    jb.setToolTipText("Next selected");
    // FIXME:
    // jb.addActionListener(nextSelectedScenario);
    // setTarget( ... );
    gbl2.setConstraints(jb, c);
    jp.add(jb);

    lowerPanel.add(jp);

    // Create a sub-panel with "Copy & Add" and "Delete Selected"
    gbl2 = new GridBagLayout();
    jp = new JPanel(gbl2);

    c.gridwidth = GridBagConstraints.REMAINDER;
    c.fill = GridBagConstraints.HORIZONTAL;
    c.anchor = GridBagConstraints.EAST;
    c.weightx = 0.0;
    jb = new JButton("Copy & Add");
    copyAndAdd = new AbstractAction() {
	public void actionPerformed(ActionEvent ae) {
	    String sname = (String)scenarioList.getSelectedValue();
	    Scenario s = sset.getScenario( sname );
	    Scenario newS = (Scenario)s.clone();
	    newS.setName( newS.getName() + "-clone" );
	    sset.addScenario( newS );
	    setTarget( newS );
	    scenarioList.setSelectedValue( newS.getName(),
					   true );
	}
    };
    jb.addActionListener(copyAndAdd);
    gbl2.setConstraints(jb, c);
    jp.add(jb);

    jb = new JButton("Delete Selected");
    deleteSelected = new AbstractAction() {
	public void actionPerformed(ActionEvent ae) {
	    String sname = (String)scenarioList.getSelectedValue();
	    int idx = scenarioList.getSelectedIndex();
	    idx =
	    ( idx-1 <= sset.scenarios.getSize()-1 ) ? idx-1 : idx;
	    Scenario sc = sset.getScenario( sname );
	    if ( sc != null) {
		sset.removeScenario( sc );
	    }
	    scenarioList.setSelectedIndex( idx );
	}
    };
    jb.addActionListener(deleteSelected);
    gbl2.setConstraints(jb, c);
    jp.add(jb);

    lowerPanel.add(jp);

    c.fill = GridBagConstraints.HORIZONTAL;
    c.anchor = GridBagConstraints.CENTER;
    gbl.setConstraints(lowerPanel, c);
    cp.add(lowerPanel);
    
    pack();
    setVisible(true);

    if (scen == null) {
	Enumeration e = sset.getAllScenarios();
	if (e.hasMoreElements()) {
	    scenarioList.
		setSelectedValue(((Scenario)e.nextElement()).
				 getName(), true);
	}
    }
    
  }

    /**
     * Implementation of ListSelectionListener.valueChanged.  When the
     * value in the list selection does change, we need to retarget
     * the ScenarioViewer.
     *
     * @param     lse The ListSelectionEvent we're interested in.
     **/
    public void valueChanged( ListSelectionEvent lse )
    {
	// Don't respond to adjusting events.
	if ( lse.getValueIsAdjusting() ) { return; }
	String scenarioName =
	    (String)scenarioList.getSelectedValue();
	if ( scenarioName == null ) {
	    return;
	}

	// Stop the current cell editing
	if ( labelTable.isEditing() ) {
	    labelTable.getCellEditor().stopCellEditing();
	}
	Scenario s = sset.getScenario( scenarioName );
	if (s == null) {
	    // FIXME: error
	} else {
	    setTarget( s );
	}
    }


    /**
     * Retargets the SceanrioViewer to view the given Scenario object.
     * This method changes the appropriate parts of the window to view
     * the data associated with the given Scenario, rather than whatever
     * it was before.
     *
     * @param     s The Scenario to view in this ScenarioViewer.
     **/
    public void setTarget(Scenario s)
    {
	if (s == null) {
	    // FIXME
	    return;
	} 

	// If the name of the currently-targeted scenario has changed,
	// then we need to update it.
	if ( sc != null ) {
	    if ( ! sc.getName().equals( nameTF.getText() ) ) {
		sset.removeScenario( sc );
		sc.setName( nameTF.getText() );
		sset.addScenario( sc );
	    }
	} // end if

	sc = s;

	/* update scenario name */
	nameTF.setText( sc.getName() );

	Component[] cs = fstructs.getComponents();
	Fstruct[] fs = sc.getFstructs();
	if ( cs.length != fs.length ) {
	    // FIXME: error
	    System.out.println("Fstruct count mismatch: " +
			       cs.length + " components, " +
			       fs.length + " fstructs.");
	} else {
	    // Retarget Fstructs
	    System.out.println( "About to retarget fstruct" );
	    for (int i=0; i<cs.length; i++) {
		System.out.println("  Fstruct " + i + "/" + cs.length);
		((FstructViewer)cs[i]).setTarget( fs[i] );
	    }
	    System.out.println("Done retargeting fsructs");
	    fstructs.validate();
	}
	// Update the table. [Fire a table data change.]
	stm.update();
    }


    /**
     * Parent AbstractAction for all Action which perform changes in
     * the selected Scenario.  This boils down to stopping table
     * editing so that the current cell editor doesn't influence the
     * other Scenario's label.
     **/
    protected class SelectionAction
	extends AbstractAction
    {
	public void actionPerformed(ActionEvent ae) {
	    if ( labelTable.isEditing() ) {
		labelTable.getCellEditor().stopCellEditing();
	    }
	}
    }


    /**
     * A TableModel for the Scenarios.  This model gets the table data
     * from the currently-selected Scenario.
     **/
    class ScenarioTableModel
	extends AbstractTableModel
    {
	public int getColumnCount() { return 3; }
	public int getRowCount() {
	    // Return the number of languages present in the ScenarioSet
	    return sset.getLanguages().size();
	}
	public String getColumnName(int row) {
	    String toRet = null;
	    switch (row) {
		case 0:
		toRet = "Toggle";
		break;
		case 1:
		toRet = "Language";
		break;
		case 2:
		toRet = "Label";
		break;
	    }
	    return toRet;
	}
	public Class getColumnClass(int col) {
	    Class toRet = null;
	    switch (col) {
	    case 0:
		toRet = java.lang.Boolean.class;
		break;
	    case 1:
	    case 2:
		toRet = java.lang.String.class;
		break;
	    } // end switch
	    return toRet;
	}
	public Object getValueAt(int row, int col) {
	    Object toRet = null;
	    String lang = (String)sset.getLanguages().elementAt( row );
	    switch (col) {
		case 0:
		    toRet = langLabelDisplay.elementAt( row );
		    break;
		case 1:
		// FIXME: If row = 0, this is the default language.  If row >
		// 0, it's the rest of the langs.
		    toRet = lang;
		    break;
		case 2:
		// FIXME: If row = 0, this is the default language label.  If
		// row > 0, it's a different lang label.

		    if (sc == null) {
			toRet = new String("No selected scenario");
		    } else {
			if ( ! ((Boolean)langLabelDisplay.
				elementAt( row )).booleanValue() ) {
			    // If toggled off, then don't display.
			    toRet = new String("(label supressed)");
			} else {
			    // FIXME: Handle the fact that if a label
			    // is unspecified, this will return
			    // null...
			    VerbComplex label = sc.getLabel( lang );
			    if (label == null) {
				toRet = " ( unlabeled ) ";
			    } else {
				toRet = label.toString();
			    }
			}
		    }
		    break;
	    } // end switch
	    return toRet;
	}
	public void setValueAt(Object o, int row, int col) {
	    switch (col) {
	      case 0:
		  // Change toggle state.  If row = 0, this is an
		  // error.
		  if (row != 0) {
		      langLabelDisplay.setElementAt( o, row );
		      fireTableStructureChanged();
		  }
		  break;
	      case 1:
		  // Error -- can't change lang name.
		  break;
	      case 2:
		  // Change label for this lang.
		  // FIXME: check to make sure that it has the appropriate
		  // number of slots filled... if not, either error or fix
		  // it [adding '-' to the end as needed].

		  // Check for "(label supressed)", and don't change
		  // anything if this is the case.
		  String lang = (String)sset.getLanguages().elementAt( row );
		  if (sc != null) {
		      VerbComplex oldLabel = (VerbComplex)sc.getLabel( lang );
		      VerbComplex newLabel = null;
		      if ( !((String)o).trim().equals("") ) {
			  newLabel = new VerbComplex( (String)o );
		      }
		      if (oldLabel != null) {
			  // Remove the old label choice
			  sset.removeLabelChoice( oldLabel.toString(), lang );
		      }
		      // Set the label
		      sc.setLabel( lang, newLabel );
		      if (newLabel != null) {
			  // Add the new label choice
			  sset.addLabelChoice( newLabel.toString(), lang );
		      }
		  }
		  break;
	  } // end switch col
      }
      public boolean isCellEditable(int row, int col) 
      {
	  if (sc == null) {
	      return false;
	  }

	  // FIXME: If a lang-label is toggled off, then it's not editable.

	  // If row = 0 and col = 0, it's the default language toggle,
	  // which is not editable.
	  // Or
	  // If col = 1, it's the language name, which is not
	  // editable.
	  if ( ((row==0) && (col==0)) || (col == 1) ) {
	      // No, we can't change the languages here.
	      return false;
	  } else {
	      if ( col == 2 ) {
		  return ((Boolean)langLabelDisplay.elementAt( row )).booleanValue();
	      } else {
		  return true;
	      }
	  }
      }
	/**
	 * Updates the table.  To be used when the currently-targeted
	 * Scenario changes.
	 **/
	public void update()
	{
	    fireTableDataChanged();
	    //fireTableStructureChanged();
	}
    }

    /**
     * The TableCellEditor impl. for the VerbComplex-containing
     * label cells.  As well, the JComboBox returned by the
     * getTableCellEditorComponent(...) method is sensitive to the row
     * number, because that controls the language which the entries in
     * the ComboBox are from.
     **/
    protected class LabelTableCellEditor
	extends DefaultCellEditor
    {
	public LabelTableCellEditor()
	{
	    super( new JComboBox() );
	}

	public Component getTableCellEditorComponent( JTable tab,
						      Object value,
						      boolean isSelected,
						      int row,
						      int col )
	{
	    Vector langs = sset.getLanguages();
	    String lang = (String)langs.elementAt( row );
	    JComboBox editor = (JComboBox)editorComponent;
	    //editor.removeAllItems();
	    if ( ((Boolean)langLabelDisplay.elementAt( row )).
		 booleanValue() ) {
		editor.setEditable( true );
		StringSortedHashtable ssh =
		    (StringSortedHashtable)sset.getLabelChoices( lang );
		if (ssh == null) 
		    System.out.println( "ssh is null" );
		Enumeration skeysEnum = ssh.getSortedKeys();
		if (skeysEnum == null)
		    System.out.println( "enum is null" );
		Vector labelChoices = new Vector();
		String elem;
		while ( skeysEnum.hasMoreElements() ) {
		    elem = (String)skeysEnum.nextElement();
		    labelChoices.addElement( elem );
		} // end while over sorted keys
		VectorComboBoxModel labelCBM =
		    new VectorComboBoxModel( labelChoices );
		editor.setModel( labelCBM );
		editor.setSelectedItem( value );
	    } else {
		editor.setEditable( false );
		editor.addItem( "(label supressed)" );
		editor.setSelectedItem( "(label supressed)" );
	    } // end if-else

	    return editor;
	}
    } // end inner class LabelTableCellEditor

}
